/**
 * class storing the current bandwidth usage for a specific point in time
 */
case class BandwidthUsage(node: Node,
                          executionTime: Double,
                          computeOperations: Long,
                          verticalCommunication: Map[String, Long],
                          lateralCommunication: Map[String, Long]) {
  // method leveraging as much bandwidth as possible
  // (the method returns the new bandwidth usage as well as the unsatisfied bandwidth)
  def reserveBandwidth(memory: Memory, communication: Long) = {
    // get the current usage of the memory level
    val currentUsage = verticalCommunication.getOrElse(memory.id, 0L)
    // compute the total communication time
    val communicationTime = (currentUsage + communication).toDouble / memory.memoryBandwidth
    if(communicationTime <= executionTime) {
      (
        BandwidthUsage(node,
          executionTime,
          computeOperations,
          verticalCommunication.updated(memory.id, (currentUsage + communication)),
          lateralCommunication),
        0L
      )
    } else {
      val slotLimit = (memory.memoryBandwidth * executionTime).toLong
      val remainingCommunication = communication - slotLimit + currentUsage
      assert(remainingCommunication > 0L)
      (
        BandwidthUsage(node,
          executionTime,
          computeOperations,
          verticalCommunication.updated(memory.id, slotLimit),
          lateralCommunication),
        remainingCommunication
      )
    }
  }

  def printAnalysisInformation(prefix: String) = {
    // prepare analysis information
    val computeInformation = s"compute=${computeOperations / executionTime * 1e-9} GFlops"
    val verticalBandwidthInformation = verticalCommunication.toList.map { usage =>
      if(usage._2 > 0) Some(s"vertical[${usage._1}]=${usage._2 * 8 / executionTime * 1e-9} GB/s") else None
    }.flatten
    val lateralBandwidthInformation = lateralCommunication.toList.map { usage =>
      if(usage._2 > 0) Some(s"lateral[${usage._1}]=${usage._2 * 8 / executionTime * 1e-9} GB/s") else None
    }.flatten
    val managementSummary = (computeInformation :: (verticalBandwidthInformation ::: lateralBandwidthInformation)).mkString(" ")
    // assemble an analysis string
    s"$prefix ${node.toString}: execution time=${executionTime * 1000 * 1000} us $managementSummary"
  }
}
