/**
 * class employing dynamic programming on order to find the optimal solution
 */
class OptimizerDP extends Optimizer {
  // provide a lazy evaluation wrapper
  class Lazy[A](x: => A) { lazy val value = x }

  // performs a brute force search
  def searchOptimalTiling(stencils: List[Stencil], outputs: List[String], searchConfiguration: List[(Tiling, Boolean)]) = {
    // compute sets that hold the results respectively dependencies up to respectively after the given index
    val results = stencils.scanLeft(Set.empty[String]) { (currentResults, stencil) =>
      currentResults + stencil.id
    }.toVector
    val dependencies = stencils.scanRight(outputs.toSet) { (stencil, currentDependencies) =>
      currentDependencies ++ stencil.dataDependencies.keys.toSet
    }.toVector
    // method that computes all possible stencil groups at the top of the memory hierarchy
    def initializeTopOfMemoryHierarchy(tiling: Tiling) = {
      // pre compute all stencil group variants
      Array.tabulate(stencils.size, stencils.size) { (fromIndex, toIndex) =>
        new Lazy(
          if (toIndex >= fromIndex) {
            val nodes = stencils.slice(fromIndex, toIndex + 1).toList
            val outputs = ((results(toIndex + 1) -- results(fromIndex)) intersect dependencies(toIndex + 1)).toList
            val candidateGroup = StencilGroup(tiling, nodes, outputs)
            val feasibilityGroup = StencilGroup(tiling, nodes, outputs)
            if (feasibilityGroup.isFeasible) {
              Some(candidateGroup)
            } else None
          } else None
        )
      }
    }
    // method that adds a new memory hierarchy level
    def addMemoryHierarchyLevel(candidateSpace: Array[Array[Lazy[Option[StencilGroup]]]], search: Boolean, tiling: Tiling) = {
      // method that nests candidate with a single parent group
      def computeSingleCandidate(fromIndex: Int, targetIndex: Int, toIndex: Int): Option[(StencilGroup, Double)] = {
        if (toIndex >= fromIndex && targetIndex >= toIndex) {
          // compute the head and the corresponding tail groups
          if(candidateSpace(fromIndex)(toIndex).value.isDefined) {
            val candidateGroup = candidateSpace(fromIndex)(toIndex).value.get
            val candidateTail = (toIndex + 1 to targetIndex).map { index =>
              val candidate = candidateSpace(index)(index).value.get
              StencilGroup(tiling, List(candidate), candidate.outputs)
            }.toList
            val feasibilityOutputs = ((results(targetIndex + 1) -- results(fromIndex)) intersect dependencies(targetIndex + 1)).toList
            val feasibilityGroup = StencilGroup(tiling, candidateGroup :: candidateTail, feasibilityOutputs)
            if (feasibilityGroup.isFeasible) {
              val performanceGroup = StencilGroup(tiling, List(candidateGroup), candidateGroup.outputs)
              Some((StencilGroup(tiling, List(candidateGroup), candidateGroup.outputs), performanceGroup.executionTime))
            } else None
          } else None
        } else None
      }
      // either search for different candidates or simple forward single group candidate
      if(search) {
        // perform optimization step finding the most efficient combinations for every from to index pair
        (0 until stencils.size).map { fromIndex =>
          (0 until stencils.size).map { targetIndex =>
            new Lazy(
              if (targetIndex >= fromIndex) {
                val result = (fromIndex to targetIndex).foldLeft(List.empty[Option[(StencilGroup, Double)]]) { (current, toIndex) =>
                  // either the optimal solution is a single group ranging between fromIndex and toIndex
                  val singleCandidate = computeSingleCandidate(fromIndex, targetIndex, toIndex)
                  // compute all combinations with partial optimal solutions previously computed
                  val mixedCandidates = current.zipWithIndex.map {
                    case (Some((partial, _)), index) =>
                      candidateSpace(toIndex - index)(toIndex).value match {
                        case Some(completion) =>
                          val feasibilityTail = (toIndex + 1 to targetIndex).map { index =>
                            candidateSpace(index)(index).value.get
                          }.toList
                          val feasibilityOutputs = ((results(targetIndex + 1) -- results(fromIndex)) intersect dependencies(targetIndex + 1)).toList
                          val feasibilityGroup = StencilGroup(tiling, partial.nodes ::: List(completion) ::: feasibilityTail, feasibilityOutputs)
                          val performanceOutputs = ((results(toIndex + 1) -- results(fromIndex)) intersect dependencies(toIndex + 1)).toList
                          if (feasibilityGroup.isFeasible) {
                            val performanceGroup = StencilGroup(tiling, partial.nodes ::: List(completion), performanceOutputs)
                            val result = Some((StencilGroup(tiling, partial.nodes ::: List(completion), performanceOutputs), performanceGroup.executionTime))
                            // regularly run garbage collection
                            System.gc()
                            Thread.sleep(100)
                            result
                          } else None
                        case None => None
                      }
                    case (None, index) => None
                  }
                  // combine the groups and find the optimum
                  val feasibleSolutions = (singleCandidate :: mixedCandidates).flatten
                  if (feasibleSolutions.nonEmpty) {
                    Some(feasibleSolutions.minBy(_._2)) :: current
                  } else {
                    None :: current
                  }
                }.head
                // print a status report
                result match {
                  case None => /*println(s"tiling ${tiling.id} ($fromIndex-$targetIndex): no solution found");*/ None
                  case Some((candidate, executionTime)) => /*println(s"tiling ${tiling.id} ($fromIndex-$targetIndex): executionTime = $executionTime");*/ Some(candidate)
                }
              } else None
            )
          }.toArray
        }.toArray
      } else {
        // in case we perform no search simply add a new nesting level
        Array.tabulate(stencils.size, stencils.size) { (fromIndex, toIndex) =>
          new Lazy(
            if (toIndex >= fromIndex) {
              // compute the head and the corresponding tail groups
              if (candidateSpace(fromIndex)(toIndex).value.isDefined) {
                val candidateGroup = candidateSpace(fromIndex)(toIndex).value.get
                Some(StencilGroup(tiling, List(candidateGroup), candidateGroup.outputs))
              } else None
            } else None
          )
        }
      }
    }
    // compute the initial configuration and afterwards apply
    val initialCandidates = initializeTopOfMemoryHierarchy(searchConfiguration.head._1)
    val finalCandidates = searchConfiguration.tail.foldLeft(initialCandidates) { (candidates, configuration) =>
      addMemoryHierarchyLevel(candidates, configuration._2, configuration._1)
    }
    // chose final candidate
    finalCandidates(0)(stencils.size - 1).value.get
  }
}
