/**
 * class implementing horizontal diffusion
 */
object StellaExperiment1 {
  // define sample configuration
  val tiling0 = Tiling(0, Memory.DeviceMemory, Memory.DeviceMemory, Domain(256,256,64), OF)
  val tiling1 = Tiling(1, Memory.DeviceMemory, Memory.DeviceMemory, Domain(64,4,64), OF)
  val tiling2 = Tiling(2, Memory.RegisterFile, Memory.RegisterFile, Domain(Int.MaxValue,Int.MaxValue,1), HEKInc)
  val tiling3 = Tiling(3, Memory.ThreadMemory, Memory.SharedMemory, Domain(1,1,1), HEPar)

  val searchConfiguration = List((tiling3, true), (tiling2, false), (tiling1, true), (tiling0, true))

  // define stencils
  val lap = Stencil("lap", 5, Map("in" -> AccessSet(List(Index(-1,0,0), Index(1,0,0), Index(0,-1,0), Index(0,1,0), Index(0,0,0)))))
  val fli = Stencil("fli", 1, Map("lap" -> AccessSet(List(Index(1,0,0), Index(0,0,0)))))
  val flj = Stencil("flj", 1, Map("lap" -> AccessSet(List(Index(0,1,0), Index(0,0,0)))))
  val out = Stencil("out", 5, Map("in" -> AccessSet.center, "wgt" -> AccessSet.center, "fli" -> AccessSet(List(Index(-1,0,0), Index(0,0,0))), "flj" -> AccessSet(List(Index(0,-1,0), Index(0,0,0)))))

  def runExperiment() = {
    println("Experiment1: HorizontalDiffusion")
    println("================================")

    // evaluate the performance of variant 1
    {
      val group3 = StencilGroup(tiling3, List(lap, fli, flj, out), List("out"))
      val group2 = StencilGroup(tiling2, List(group3), List("out"))
      val group1 = StencilGroup(tiling1, List(group2), List("out"))
      val group0 = StencilGroup(tiling0, List(group1), List("out"))

      println(s"-> evaluating storage requirements and performance")

      val t0 = System.nanoTime()
      val performance = group0.evaluatePerformance
      val requirements = group0.evaluateStorage
      val t1 = System.nanoTime()

      println(s"-> elapsed time: ${(t1 - t0) / 1000 / 1000} ms")

      // evaluate the execution time for a specific solution
      println("\n1) Requirements (v1)")
      println("--------------------")
      println(requirements)
      println("-> " + requirements.verifyStorageRequirements)

      println("\n2) Performance (v1)")
      println("-------------------")

      println(performance.printNodeInformation("->"))
      println(performance.printAnalysisInformation("->"))
    }

    // evaluate the performance of variant 2
    {
      val group5 = StencilGroup(tiling2, List(StencilGroup(tiling3, out)), List("out"))
      val group4 = StencilGroup(tiling2, List(StencilGroup(tiling3, flj)), List("flj"))
      val group3 = StencilGroup(tiling2, List(StencilGroup(tiling3, fli)), List("fli"))
      val group2 = StencilGroup(tiling2, List(StencilGroup(tiling3, lap)), List("lap"))

      val group1 = StencilGroup(tiling1, List(group2, group3, group4, group5), List("out"))
      val group0 = StencilGroup(tiling0, List(group1), List("out"))

      println(s"-> evaluating storage requirements and performance")

      val t0 = System.nanoTime()
      val performance = group0.evaluatePerformance
      val requirements = group0.evaluateStorage
      val t1 = System.nanoTime()

      println(s"-> elapsed time: ${(t1 - t0) / 1000 / 1000} ms")

      // evaluate the execution time for a specific solution
      println("\n1) Requirements (v2)")
      println("--------------------")
      println(requirements)
      println("-> " + requirements.verifyStorageRequirements)

      println("\n2) Performance (v2)")
      println("-------------------")

      println(performance.printNodeInformation("->"))
      println(performance.printAnalysisInformation("->"))
    }

    println("")
  }

  def runExperimentsGPU() = {
    // define sample configuration
    val tiling0 = Tiling(0, Memory.DeviceMemory, Memory.DeviceMemory, Domain(256,256,64), OF)
    val tiling1 = Tiling(1, Memory.DeviceMemory, Memory.DeviceMemory, Domain(64,4,64), OF)
    val tiling2 = Tiling(2, Memory.RegisterFile, Memory.RegisterFile, Domain(Int.MaxValue,Int.MaxValue,1), HEKInc)
    val tiling3 = Tiling(3, Memory.ThreadMemory, Memory.SharedMemory, Domain(1,1,1), HEPar)

    // no fusion
    {
      val group4 = StencilGroup(tiling1, List(StencilGroup(tiling2, List(StencilGroup(tiling3, out)), List("out"))), List("out"))
      val group3 = StencilGroup(tiling1, List(StencilGroup(tiling2, List(StencilGroup(tiling3, flj)), List("flj"))), List("flj"))
      val group2 = StencilGroup(tiling1, List(StencilGroup(tiling2, List(StencilGroup(tiling3, fli)), List("fli"))), List("fli"))
      val group1 = StencilGroup(tiling1, List(StencilGroup(tiling2, List(StencilGroup(tiling3, lap)), List("lap"))), List("lap"))
      val group0 = StencilGroup(tiling0, List(group1, group2, group3, group4), List("out"))

      println(s"exp1 (GPU, no fusion): ${group0.executionTime} ${group0.isFeasible}")
    }

    // hand-tuned and optimized
    {
      val group3 = StencilGroup(tiling3, List(lap, fli, flj, out), List("out"))
      val group2 = StencilGroup(tiling2, List(group3), List("out"))
      val group1 = StencilGroup(tiling1, List(group2), List("out"))
      val group0 = StencilGroup(tiling0, List(group1), List("out"))
      println(s"exp1 (GPU, hand-tuned): ${group0.executionTime} ${group0.isFeasible}")
      println(s"exp1 (GPU, optimized): ${group0.executionTime} ${group0.isFeasible}")
    }
  }

  def runExperimentsCPU() = {
    // define sample configuration
    val tiling0 = Tiling(0, Memory.HostMemory, Memory.HostMemory, Domain(256, 256, 64), OF)
    val tiling1 = Tiling(1, Memory.HostCache, Memory.HostCache, Domain(8, 8, 64), OF)

    // no fusion
    {
      val group4 = StencilGroup(tiling1, out)
      val group3 = StencilGroup(tiling1, flj)
      val group2 = StencilGroup(tiling1, fli)
      val group1 = StencilGroup(tiling1, lap)
      val group0 = StencilGroup(tiling0, List(group1, group2, group3, group4), List("out"))

      println(s"exp1 (CPU, no fusion): ${group0.executionTime} ${group0.isFeasible}")
    }

    // hand-tuned and optimized
    {
      val group1 = StencilGroup(tiling1, List(lap, fli, flj, out), List("out"))
      val group0 = StencilGroup(tiling0, List(group1), List("out"))
      println(s"exp1 (CPU, hand-tuned): ${group0.executionTime} ${group0.isFeasible}")
      println(s"exp1 (CPU, optimized): ${group0.executionTime} ${group0.isFeasible}")
    }
  }

  def performOptimizationGPU(optimizationRange: (Int, Int)) = {
    println("Performing Optimization GPU")
    println("---------------------------")

    // define sample configuration
    val tiling0 = Tiling(0, Memory.DeviceMemory, Memory.DeviceMemory, Domain(256,256,64), OF)
    val tiling1 = Tiling(1, Memory.DeviceMemory, Memory.DeviceMemory, Domain(64,4,64), OF)
    val tiling2 = Tiling(2, Memory.RegisterFile, Memory.RegisterFile, Domain(Int.MaxValue,Int.MaxValue,1), HEKInc)
    val tiling3 = Tiling(3, Memory.ThreadMemory, Memory.SharedMemory, Domain(1,1,1), HEPar)

    val searchConfiguration = List((tiling3, true), (tiling2, false), (tiling1, true), (tiling0, true))

    val t0 = System.nanoTime()
    val optimizer = new OptimizerDP()
    val optimum = optimizer.searchOptimalTopology(optimizationRange, List(lap, fli, flj, out), List("out"), searchConfiguration)
    val t1 = System.nanoTime()

    println(s"-> elapsed time: ${(t1 - t0) / 1000 / 1000} ms")

    val performance = optimum.evaluatePerformance
    val storage = optimum.evaluateStorage
    println(storage.prettyPrint("->"))
    println(performance.printNodeInformation("->"))
    println(performance.printAnalysisInformation("->"))

    println("")
  }

  def performOptimizationCPU(optimizationRange: (Int, Int)) = {
    println("Performing Optimization CPU")
    println("---------------------------")

    // search the optimal CPU solution
    val tiling0 = Tiling(0, Memory.HostMemory, Memory.HostMemory, Domain(256, 256, 64), OF)
    val tiling1 = Tiling(1, Memory.HostCache, Memory.HostCache, Domain(8, 8, 64), OF)
    val searchConfiguration = List((tiling1, true), (tiling0, true))

    val t0 = System.nanoTime()
    val optimizer = new OptimizerDP()
    val optimum = optimizer.searchOptimalTopology(optimizationRange, List(lap, fli, flj, out), List("out"), searchConfiguration)
    val t1 = System.nanoTime()

    println(s"-> elapsed time: ${(t1 - t0) / 1000 / 1000} ms")

    val performance = optimum.evaluatePerformance
    val storage = optimum.evaluateStorage
    println(storage.prettyPrint("->"))
    println(performance.printNodeInformation("->"))
    println(performance.printAnalysisInformation("->"))

    println("")
  }
}
