
/**
 * a group represents a group of stencils executed together
 */
object StencilGroup {
  def apply(tiling: Tiling, stencil: Stencil): StencilGroup = {
    StencilGroup(tiling, List(stencil), stencil.outputs)
  }
}

case class StencilGroup(tiling: Tiling, nodes: List[Node], outputs: List[String]) extends Node {
  // check the output is computed by the group
  assert((outputs.toSet -- nodes.flatMap(_.outputs).toSet).isEmpty)

  // implement the node interface
  val id = outputs.mkString("&")

  // compute all dependencies
  val inputNodes = nodes.flatMap(_.inputs).toSet
  val outputNodes = nodes.flatMap(_.outputs).toSet

  // compute the group inputs and all internal dependencies
  val inputs = (inputNodes -- outputNodes).toList
  val temporaries = (inputNodes intersect outputNodes).toList

  // method that computes a flattened node list
  val stencils: List[Stencil] = nodes.flatMap {
    case group: StencilGroup => group.stencils
    case stencil: Stencil => List(stencil)
  }

  // compute two dependency maps considering the node dependencies and all nested stencil dependencies respectively
  val dependencies = {
    val initial = isl.UnionMap.empty(isl.Space.alloc(ISLContext.ctx, 0, 3, 3))
    val union = nodes.foldLeft(initial) { (current, node) => current union node.dependencies }
    val closure = union.transitiveClosure(Array(0))
    val space = isl.Space.setAlloc(ISLContext.ctx, 0, 3)
    val inputSet = inputs.foldLeft(isl.UnionSet.empty(space)) { (current, id) =>
      current.addSet(isl.Set.universe(space).setTupleId(ISLContext.findOrCreateID(id)))
    }
    val outputSet = outputs.foldLeft(isl.UnionSet.empty(space)) { (current, id) =>
      current.addSet(isl.Set.universe(space).setTupleId(ISLContext.findOrCreateID(id)))
    }
    closure.intersectDomain(outputSet).intersectRange(inputSet).coalesce()
  }
  val flattenDependencies = {
    val initial = isl.UnionMap.empty(isl.Space.alloc(ISLContext.ctx, 0, 3, 3))
    val union = nodes.foldLeft(initial) { (current, node) => current union node.flattenDependencies }
    union.transitiveClosure(Array(0)).coalesce()
  }

  // performance properties used for optimization
  lazy val executionTime = evaluatePerformance.executionTime
  lazy val isFeasible = evaluateStorage.verifyStorageRequirements && nodes.nonEmpty

  // method that computes that extends the tiling with the stencil group tiling hierarchy
  def computeTilingImpl(parentOutputs: List[String], parentShifts: Map[String, List[Int]], parentTiling: isl.UnionMap, outputTiling: isl.UnionMap) = {
    import Conversions._
    // compute potential stencil shifts for the stencil group
    val stencilShifts = tiling.haloStrategy match {
      case OF => Map.empty[String, List[Int]]
      case HEPar => parentShifts
      case HEKInc =>
        val outputDependencies = outputs.map { id => flattenDependencies.intersectDomain(computeStencilFilter(id)) }
        // search for each stencil output pair the maximal offset using the delta functionality
        (inputs ::: stencils.map(_.id)).map { id =>
          val shifts = parentShifts.getOrElse(id, List(0, 0, 0))
          val maxOffset = outputDependencies.map { dependencies =>
            val stencilDependencies = dependencies.intersectRange(computeStencilFilter(id))
            var offset = shifts(2)
            val searchMaximalOffset: isl.Map => Unit = { dependencyMap =>
              val resetMap = dependencyMap.resetTupleId(isl.DimType.In).resetTupleId(isl.DimType.Out)
              val samplePoint = resetMap.deltas().projectOut(isl.DimType.Set, 0, 2).lexmax().samplePoint()
              offset = Math.max(offset, samplePoint.getCoordinateVal(isl.DimType.Set, 0).toString().toInt)
            }
            stencilDependencies.foreachMap(searchMaximalOffset)
            offset
          }.max
          (id, List(shifts(0), shifts(1), maxOffset))
        }.toMap
    }
    // perform a halo exchange tiling on the given nodes
    def performHaloExchangeTiling(nodes: List[String], parentTiling: isl.UnionMap) = {
      // compute a tiling expression for the given stencil id
      def computeTilingExpression(id: String) = {
        // tile all output stencils and compute the range product with the parent tiling
        val space = isl.Space.alloc(ISLContext.ctx, 0, 3, 3)
        // compute a halo exchange tiling map used of the output respectively all nodes depending on the halo strategy
        val sizes = List(tiling.tileSize.iSize, tiling.tileSize.jSize, tiling.tileSize.kSize)
        val shifts = stencilShifts.getOrElse(id, List(0, 0, 0))
        val initial = isl.AffList.alloc(ISLContext.ctx, 0)
        val expressions = (0 to 2).foldLeft(initial) { (current, index) =>
          if (sizes(index) == Int.MaxValue) {
            // map the dimension to zero in case the tile size is inf
            val localSpace = isl.LocalSpace.fromSpace(space.domain())
            current.add(isl.Aff.valOnDomain(localSpace, isl.Val.fromInt(ISLContext.ctx, 0)))
          } else {
            // otherwise divide the dimension index by the tile size
            val localSpace = isl.LocalSpace.fromSpace(space.domain())
            val indexVariable = isl.Aff.varOnDomain(localSpace, isl.DimType.Set, index)
            val tileSize = isl.Aff.valOnDomain(localSpace, isl.Val.fromInt(ISLContext.ctx, sizes(index)))
            val tileShift = isl.Aff.valOnDomain(localSpace, isl.Val.fromInt(ISLContext.ctx, shifts(index)))
            val expression = indexVariable.sub(tileShift).div(tileSize).floor()
            current.add(expression)
          }
        }
        // convert affine expressions to a tiling map and apply the tiling to the individual nodes
        val tilingMap = isl.Map.fromMultiAff(isl.MultiAff.fromAffList(space, expressions))
        tilingMap.setTupleId(isl.DimType.In, ISLContext.findOrCreateID(id))
      }
      // compute union of the individual node tilings
      val nodeTiling = nodes.map { id =>
        isl.UnionMap.fromMap(computeTilingExpression(id))
      }.reduce(_.union(_))
      // finally combine the node tiling with the parent tiling
      parentTiling.flatRangeProduct(nodeTiling)
    }
    // do nothing in case there are no nested nodes
    if (nodes.isEmpty) {
      List((parentTiling, outputTiling))
    } else {
      // compute a halo exchange tiling for all outputs
      val initial = performHaloExchangeTiling(parentOutputs, outputTiling)
      // apply the different tiling strategies node by node
      val tilings = nodes.scanRight(initial) { (node, current) =>
        tiling.haloStrategy match {
          case OF =>
            current union current.applyDomain(node.flattenDependencies)
          case _ =>
            val dependencies = node.inputs ::: node.stencils.map(_.id)
            current union performHaloExchangeTiling(dependencies, parentTiling)
        }
      }.sliding(2).toList
      // for each interior node compute a new output tiling only containing the outputs of the corresponding node
      val interiorTilings = (nodes zip tilings).dropRight(1).map { element =>
        val outputFilter = element._1.outputs.map { id => isl.UnionSet.fromSet(computeStencilFilter(id)) }.reduce(_.union(_))
        val outputTiling = element._2(1).intersectDomain(outputFilter)
        element._1.computeTilingImpl(element._1.outputs, stencilShifts, element._2(0), outputTiling)
      }.flatten
      // for the last node forward all outputs of the parent outputs
      val lastTiling = nodes.last.computeTilingImpl(parentOutputs, stencilShifts, tilings.last(0), tilings.last(1))
      interiorTilings ::: lastTiling
    }
  }

  // method that computes the stencil group tiling
  def computeTiling = {
    // compute an empty tiling map containing all stencils of the stencil group
    val space = isl.Space.alloc(ISLContext.ctx, 0, 3, 0)
    val emptyTiling = stencils.foldLeft(isl.UnionMap.empty(space)) { (current, stencil) =>
      val stencilMap = isl.Map.universe(space).setTupleId(isl.DimType.In, ISLContext.findOrCreateID(stencil.id))
      current union isl.UnionMap.fromMap(stencilMap)
    }
    // compute the actual stencil group tiling
    val (stencilTiling, outputTiling) = computeTilingImpl(outputs, Map.empty[String, List[Int]], emptyTiling, emptyTiling).unzip
    // compute two maps that associate the stencil id
    ((stencils.map(_.id) zip stencilTiling).toMap, (stencils.map(_.id) zip outputTiling).toMap)
  }

  // method that tiles all stencil inputs given the stencil tiling
  def computeInputTiling(stencilTiling: Map[String, isl.UnionMap]) = {
    stencilTiling.map { case (id, tiling) =>
      val dependencies = stencils.find(stencil => stencil.id == id).get.dependencies
      (id, tiling.applyDomain(dependencies))
    }
  }

  // method that computes a stencil set given the stencil id
  def computeStencilFilter(id: String) = {
    isl.Set.universe(isl.Space.setAlloc(ISLContext.ctx, 0, 3)).setTupleId(ISLContext.findOrCreateID(id))
  }

  // method that computes a stencil set given the stencil id
  def computeTileFilter(fixingLevel: Int, stencilTiling: isl.UnionMap) = {
    assert(!stencilTiling.isEmpty)
    // fix all dimensions up to the fixing level
    (0 until fixingLevel * 3).foldLeft(isl.Set.fromUnionSet(stencilTiling.range())) { (current, dimension) =>
      current.fixVal(isl.DimType.Set, dimension, isl.Val.fromInt(ISLContext.ctx, 0))
    }
  }

  // method that projects the tiling of a stencil to a given tiling level
  def projectStencilTiling(id: String, fixingLevel: Int, projectionLevel: Int, tilingMap: isl.UnionMap) = {
    val stencilFilter = computeStencilFilter(id)
    val stencilTiling = tilingMap.intersectDomain(stencilFilter)
    val tilingFilter = computeTileFilter(fixingLevel, stencilTiling)
    // project out all dimensions above the tiling level
    val remainder = tilingFilter.getSpace.dim(isl.DimType.Set) - projectionLevel * 3
    isl.Map.fromUnionMap(stencilTiling.intersectRange(tilingFilter)).projectOut(isl.DimType.Out, projectionLevel * 3, remainder)
  }

  // method computing the stencil group storage requirements
  def computeStorage(tilingLevel: Int, inputTiling: Map[String, isl.UnionMap]): StorageRequirements = {
    // compute the union of all stencil inputs tilings
    val initial = isl.UnionMap.empty(isl.Space.alloc(ISLContext.ctx, 0, 3, 3))
    val inputs = stencils.map { stencil =>
      stencil.inputs.map { id =>
        isl.UnionMap.fromMap(projectStencilTiling(id, tilingLevel, tilingLevel, inputTiling(stencil.id)))
      }
    }.flatten.foldLeft(initial)(_.union(_))
    // iterate over all node inputs and extract the corresponding domain size
    val temporaryStorage = nodes.map(_.inputs).flatten.distinct.map { id =>
      (id, Barvinok.count(inputs.intersectDomain(computeStencilFilter(id)).wrap().toString).toLong)
    }.toMap
    // compute the storage requirement for all input dependencies
    StorageRequirements(
      this,
      temporaryStorage,
      nodes.map {
        case node: StencilGroup => Some(node.computeStorage(tilingLevel + 1, inputTiling))
        case _ => None
      }
    )
  }

  // evaluate the storage requirements of a stencil group
  def evaluateStorage = {
    // compute the storage requirements recursively using the input tiling
    val (stencilTiling, _) = computeTiling
    computeStorage(1, computeInputTiling(stencilTiling))
  }

  // method computing the stencil group storage requirements
  def computeBandwidth(tilingLevel: Int,
                       inputTiling: Map[String, isl.UnionMap],
                       stencilTiling: Map[String, isl.UnionMap],
                       outputTiling: Map[String, isl.UnionMap]): BandwidthRequirements = {
    // evaluate the compute intensity of nested stencils
    val computationalIntensity = nodes.map {
      case node: Stencil =>
        val tileFilter = computeTileFilter(1, stencilTiling(node.id))
        val applications = Barvinok.count(stencilTiling(node.id).intersectRange(tileFilter).wrap().toString).toLong
        Some(applications * node.computeOperations)
      case _ => None
    }
    // iterate over all nodes and compute the vertical communication node by node
    val verticalCommunication = nodes.map { node =>
      // extract the input stencils
      val readTiling = node.stencils.map { stencil => inputTiling(stencil.id) }.reduce(_.union(_))
      val writeTiling = outputTiling(node.stencils.last.id)
      // compute the number of loads and stores
      val reads = node.inputs.map { id =>
        val projectedInputs = projectStencilTiling(id, 1, tilingLevel, readTiling)
        (id, Barvinok.count(projectedInputs.wrap().toString).toLong)
      }.toMap
      val writes = node.outputs.map { id =>
        val projectedOutputs = projectStencilTiling(id, 1, tilingLevel, writeTiling)
        (id, Barvinok.count(projectedOutputs.wrap().toString).toLong)
      }.toMap
      reads ++ writes
    }
    // compare the read and write sets of all temporary values in order to evaluate the amount of lateral communication
    val temporaryCommunication = temporaries.map { id =>
      // compute the output node and the input nodes potentially consuming a temporary
      val outputNode = nodes.find(node => node.stencils.map(_.id).contains(id)).get
      val inputNodes = nodes.dropWhile(node => node != outputNode).tail
      // compute the input and output sets for the given stencil
      val inputStencils = inputNodes.flatMap(_.stencils.map(_.id))
      val communication = (1 to tilingLevel).map { level =>
        val projectedInputs = projectStencilTiling(id, 1, level, inputStencils.map(inputTiling(_)).reduce(_.union(_)))
        val projectedOutputs = projectStencilTiling(id, 1, level, outputTiling(outputNode.stencils.last.id))
        // define the lateral communication as set difference
        val communication = projectedInputs.subtract(projectedOutputs).wrap()
        if (communication.isEmpty) 0L else Barvinok.count(communication.toString).toLong
      }.toList
      // correct the lateral communication by removing the communication handled on lower levels
      (id, (communication zip communication.scanLeft(0L)(_ + _)).map { case (com, sum) => com - sum })
    }.toMap
    // associate the temporary communication to the individual nodes
    val lateralCommunication = nodes.map { node =>
      node.outputs.map { id =>
        temporaryCommunication.get(id) match {
          case Some(communication) if communication.sum > 0L => Some((id, communication))
          case _ => None
        }
      }.flatten.toMap
    }
    // compute the nested bandwidth requirements recursively and return
    BandwidthRequirements(
      this,
      computationalIntensity,
      verticalCommunication,
      lateralCommunication,
      nodes.map {
        case node: StencilGroup => Some(node.computeBandwidth(tilingLevel + 1, inputTiling, stencilTiling, outputTiling))
        case _ => None
      }
    )
  }

  // evaluate the performance of a stencil group
  def evaluatePerformance = {
    // compute bandwidth and compute intensity recursively using all tiling information
    val (stencilTiling, outputTiling) = computeTiling
    computeBandwidth(1, computeInputTiling(stencilTiling), stencilTiling, outputTiling)
  }

  override def toString = { s"[${nodes.mkString(",")}]" }
}

//  // compute the storage requirement for a single domain considering the maximal active set
//  def computeActiveTemporaryStorageOld(storageUsage: Map[String, AccessSet]) = {
//    // compute the active set
//    val touchedAggregation = nodes.scanLeft(Set.empty[String]) { (current, node) => current ++ (node.id :: node.inputs) }.tail
//    val inputAggregation = nodes.scanRight(Set.empty[String]) { (node, current) => current ++ node.inputs }.dropRight(1)
//    val activeTemporaries = (touchedAggregation zip inputAggregation).map(x => x._1 intersect x._2)
//    // compute for each element the required storage
//    val storageSizes = activeTemporaries.map { temporary =>
//      temporary.foldLeft(0L) { (cardinality, id) =>
//        val currentUsage = storageUsage.getOrElse(id, AccessSet.empty)
//        val temporaryStorage = currentUsage.cardinality(tiling.tileSize)
//        cardinality + temporaryStorage
//      }
//    }
//    if(storageSizes.isEmpty) 0L
//    else storageSizes.max
//  }
