import org.scalatest._

/**
 * test the cardinality computation
 */
class ComputeExecutionTimeSpec extends FlatSpec with Matchers {
  val lap = Stencil("lap", 5, Map("in" -> AccessSet(List(Index(-1, 0, 0), Index(1, 0, 0), Index(0, -1, 0), Index(0, 1, 0), Index(0, 0, 0)))))
  val fli = Stencil("fli", 1, Map("lap" -> AccessSet(List(Index(1, 0, 0), Index(0, 0, 0)))))
  val flj = Stencil("flj", 1, Map("lap" -> AccessSet(List(Index(0, 1, 0), Index(0, 0, 0)))))
  val out = Stencil("out", 5, Map("in" -> AccessSet.center, "wgt" -> AccessSet.center, "fli" -> AccessSet(List(Index(-1, 0, 0), Index(0, 0, 0))), "flj" -> AccessSet(List(Index(0, -1, 0), Index(0, 0, 0)))))

  "a StencilGroup" should "should compute the OF bandwdith usage" in {
    val tiling = Tiling(0, Memory.DeviceMemory, Memory.DeviceMemory, Domain(32, 32, 32), OF)
    val stencilGroup = StencilGroup(tiling, List(lap, fli, flj, out), List("out"))

    val performance = stencilGroup.evaluatePerformance
    val executionTime = performance.executionTime
    val bandwidthUsage = performance.computeBandwidthUsage(List.empty[Tiling])

    bandwidthUsage should have size 4
    executionTime should be (bandwidthUsage.map(_.executionTime).sum)

    // make sure the bandwidth and compute values are summed up correctly
    val computation = performance.computeOperations(0)
    val communication = performance.totalVerticalCommunication(0)
    "lap" should be (bandwidthUsage(0).node.id)
    computation.get should be (bandwidthUsage(0).computeOperations)
    communication should be (bandwidthUsage(0).verticalCommunication.values.sum)

    // make sure the execution time is computed properly
    communication / Memory.DeviceMemory.memoryBandwidth should be (bandwidthUsage(0).executionTime)
  }

  "a StencilGroup" should "should compute the OF>OF>OF bandwidth usage" in {
    val tiling0 = Tiling(0, Memory.DeviceMemory, Memory.DeviceMemory, Domain(8, 8, 8), OF)
    val tiling1 = Tiling(1, Memory.DeviceMemory, Memory.DeviceMemory, Domain(4, 4, 4), OF)
    val tiling2 = Tiling(2, Memory.SharedMemory, Memory.DeviceMemory, Domain(1, 1, 1), OF)

    val lap = Stencil("lap", 5, Map("in" -> AccessSet(List(Index(-1,0,0), Index(1,0,0), Index(0,-1,0), Index(0,1,0), Index(0,0,0)))))
    val fli = Stencil("fli", 1, Map("lap" -> AccessSet(List(Index(1,0,0), Index(0,0,0)))))
    val flj = Stencil("flj", 1, Map("lap" -> AccessSet(List(Index(0,1,0), Index(0,0,0)))))
    val out = Stencil("out", 5, Map("in" -> AccessSet.center, "wgt" -> AccessSet.center, "fli" -> AccessSet(List(Index(-1,0,0), Index(0,0,0))), "flj" -> AccessSet(List(Index(0,-1,0), Index(0,0,0)))))
    val group1 = StencilGroup(tiling2, List(lap, fli, flj, out), List("out"))
    val group2 = StencilGroup(tiling1, List(group1), List("out"))
    val stencilGroup = StencilGroup(tiling0, List(group2), List("out"))

    val performance = stencilGroup.evaluatePerformance
    val executionTime = performance.executionTime
    val bandwidthUsage = performance.computeBandwidthUsage(List.empty[Tiling])

    bandwidthUsage should have size 5
    executionTime should be (bandwidthUsage.map(_.executionTime).sum)

    // verify the device memory communication
    performance.nestedRequirements(0).get.totalVerticalCommunication(0) should be (bandwidthUsage.map(_.verticalCommunication("DM")).sum)
    performance.nestedRequirements(0).get.totalVerticalCommunication(0) / Memory.DeviceMemory.memoryBandwidth should be (executionTime +- 1e-10)

    // verify the shared memory communication
    performance.nestedRequirements(0).get.nestedRequirements(0).get.totalVerticalCommunication(0) should be (bandwidthUsage(0).verticalCommunication("SM"))
    performance.nestedRequirements(0).get.nestedRequirements(0).get.computeOperations(0).get should be (bandwidthUsage(0).computeOperations)
    performance.nestedRequirements(0).get.nestedRequirements(0).get.totalVerticalCommunication(0) / Memory.SharedMemory.memoryBandwidth should be (bandwidthUsage(0).executionTime)
    bandwidthUsage(0).verticalCommunication("DM") / Memory.DeviceMemory.memoryBandwidth should be (bandwidthUsage(0).executionTime +- 1e-10)
  }

  "a StencilGroup" should "should compute the HE>HE>HE bandwidth usage" in {
    val tiling0 = Tiling(0, Memory.RegisterFile, Memory.DeviceMemory, Domain(8, 8, 8), HEPar)
    val tiling1 = Tiling(1, Memory.RegisterFile, Memory.SharedMemory, Domain(4, 4, 4), HEPar)
    val tiling2 = Tiling(2, Memory.RegisterFile, Memory.RegisterFile, Domain(1, 1, 1), HEPar)

    val lap = Stencil("lap", 5, Map("in" -> AccessSet(List(Index(-1, 0, 0), Index(1, 0, 0), Index(0, -1, 0), Index(0, 1, 0), Index(0, 0, 0)))))
    val fli = Stencil("fli", 1, Map("lap" -> AccessSet(List(Index(1, 0, 0), Index(0, 0, 0)))))
    val flj = Stencil("flj", 1, Map("lap" -> AccessSet(List(Index(0, 1, 0), Index(0, 0, 0)))))
    val out = Stencil("out", 5, Map("in" -> AccessSet.center, "wgt" -> AccessSet.center, "fli" -> AccessSet(List(Index(-1, 0, 0), Index(0, 0, 0))), "flj" -> AccessSet(List(Index(0, -1, 0), Index(0, 0, 0)))))
    val group1 = StencilGroup(tiling2, List(lap, fli, flj, out), List("out"))
    val group2 = StencilGroup(tiling1, List(group1), List("out"))
    val stencilGroup = StencilGroup(tiling0, List(group2), List("out"))

    val performance = stencilGroup.evaluatePerformance
    val executionTime = performance.executionTime
    val bandwidthUsage = performance.computeBandwidthUsage(List.empty[Tiling])

    bandwidthUsage should have size 4
    executionTime should be(bandwidthUsage.map(_.executionTime).sum)

    // verify the lateral communication is considered for the execution time estimation
    val expectedCommunication = performance.nestedRequirements(0).get.nestedRequirements(0).get.totalLateralCommunication
    val lapCommunication =
      bandwidthUsage(0).lateralCommunication("DM") ::
        bandwidthUsage(0).lateralCommunication("SM") ::
        bandwidthUsage(0).lateralCommunication("RF") :: Nil
    expectedCommunication(0) should be (lapCommunication)
    val fliCommunication =
      bandwidthUsage(1).lateralCommunication("DM") ::
        bandwidthUsage(1).lateralCommunication("SM") ::
        bandwidthUsage(1).lateralCommunication("RF") :: Nil
    expectedCommunication(1) should be (fliCommunication)
    val fljCommunication =
      bandwidthUsage(2).lateralCommunication("DM") ::
        bandwidthUsage(2).lateralCommunication("SM") ::
        bandwidthUsage(2).lateralCommunication("RF") :: Nil
    expectedCommunication(2) should be (fljCommunication)
    val outCommunication =
      bandwidthUsage(3).lateralCommunication("DM") ::
        bandwidthUsage(3).lateralCommunication("SM") ::
        bandwidthUsage(3).lateralCommunication("RF") :: Nil
    List(0,0,0) should be (outCommunication)

    // verify the execution time is dominated by the lateral communication
    val lapExecutionTime = List(
      bandwidthUsage(0).lateralCommunication("DM") / Memory.DeviceMemory.memoryBandwidth,
      bandwidthUsage(0).lateralCommunication("SM") / Memory.SharedMemory.memoryBandwidth,
      bandwidthUsage(0).lateralCommunication("RF") / Memory.RegisterFile.memoryBandwidth
    ).max
    lapExecutionTime should be (bandwidthUsage(0).executionTime)
  }

}
