import org.scalatest._

/**
 * test the performance evaluation
 */
class EvaluatePerformanceSpec extends FlatSpec with Matchers {

  "a StencilGroup" should "compute the OF bandwidth" in {
    val tiling0 = Tiling(0, Memory.HostMemory, Memory.HostMemory, Domain(5, 2, 1), OF)
    val lap = Stencil("lap", 5, Map("in" -> AccessSet(List(Index(-1,0,0), Index(1,0,0), Index(0,-1,0), Index(0,1,0), Index(0,0,0)))))
    val fli = Stencil("fli", 1, Map("lap" -> AccessSet(List(Index(1,0,0), Index(0,0,0)))))
    val flj = Stencil("flj", 1, Map("lap" -> AccessSet(List(Index(0,1,0), Index(0,0,0)))))
    val out = Stencil("out", 5, Map("in" -> AccessSet.center, "wgt" -> AccessSet.center, "fli" -> AccessSet(List(Index(-1,0,0), Index(0,0,0))), "flj" -> AccessSet(List(Index(0,-1,0), Index(0,0,0)))))
    val group = StencilGroup(tiling0, List(lap, fli, flj, out), List("out"))

    // compute evaluation domain
    val performance = group.evaluatePerformance

    // verify the computation intensity of lap. fli, flj, out
    performance.computeOperations(0) should be (Some (5 * (5 * 2 + 2 * 5  + 2 * 2)))
    performance.computeOperations(1) should be (Some (1 * (5 * 2 + 1 * 2)))
    performance.computeOperations(2) should be (Some (1 * (5 * 2 + 1 * 5)))
    performance.computeOperations(3) should be (Some (5 * (5 * 2)))

    // verify the vertical communication of lap, fli, flj, out
    performance.verticalCommunication(0) should be (Map("in" -> 42, "lap" -> 24))
    performance.verticalCommunication(1) should be (Map("lap" -> 14, "fli" -> 12))
    performance.verticalCommunication(2) should be (Map("lap" -> 20, "flj" -> 15))
    performance.verticalCommunication(3) should be (Map("fli" -> 12, "flj" -> 15, "in" -> 10, "wgt" -> 10, "out" -> 10))

    // verify the lateral communication of lap, fli, flj, out
    performance.lateralCommunication(0) should be (Map())
    performance.lateralCommunication(1) should be (Map())
    performance.lateralCommunication(2) should be (Map())
    performance.lateralCommunication(3) should be (Map())
  }

  "a StencilGroup" should "compute the HE bandwidth" in {
    val tiling0 = Tiling(0, Memory.HostMemory, Memory.HostMemory, Domain(5, 2, 1), HEPar)
    val lap = Stencil("lap", 5, Map("in" -> AccessSet(List(Index(-1,0,0), Index(1,0,0), Index(0,-1,0), Index(0,1,0), Index(0,0,0)))))
    val fli = Stencil("fli", 1, Map("lap" -> AccessSet(List(Index(1,0,0), Index(0,0,0)))))
    val flj = Stencil("flj", 1, Map("lap" -> AccessSet(List(Index(0,1,0), Index(0,0,0)))))
    val out = Stencil("out", 5, Map("in" -> AccessSet.center, "wgt" -> AccessSet.center, "fli" -> AccessSet(List(Index(-1,0,0), Index(0,0,0))), "flj" -> AccessSet(List(Index(0,-1,0), Index(0,0,0)))))
    val group = StencilGroup(tiling0, List(lap, fli, flj, out), List("out"))

    // compute evaluation domain
    val performance = group.evaluatePerformance

    // verify the computation intensity of lap. fli, flj, out
    performance.computeOperations(0) should be (Some (5 * (5 * 2)))
    performance.computeOperations(1) should be (Some (1 * (5 * 2)))
    performance.computeOperations(2) should be (Some (1 * (5 * 2)))
    performance.computeOperations(3) should be (Some (5 * (5 * 2)))

    // verify the vertical communication of lap, fli, flj, out
    performance.verticalCommunication(0) should be (Map("in" -> 24, "lap" -> 10))
    performance.verticalCommunication(1) should be (Map("lap" -> 12, "fli" -> 10))
    performance.verticalCommunication(2) should be (Map("lap" -> 15, "flj" -> 10))
    performance.verticalCommunication(3) should be (Map("fli" -> 12, "flj" -> 15, "in" -> 10, "wgt" -> 10, "out" -> 10))

    // verify the lateral communication of lap, fli, flj, out
    performance.lateralCommunication(0) should be (Map("lap" -> List(7)))
    performance.lateralCommunication(1) should be (Map("fli" -> List(2)))
    performance.lateralCommunication(2) should be (Map("flj" -> List(5)))
    performance.lateralCommunication(3) should be (Map())
  }

  "a StencilGroup" should "compute the OF>OF bandwidth" in {
    val tiling0 = Tiling(0, Memory.HostMemory, Memory.HostMemory, Domain(4, 4, 4), OF)
    val tiling1 = Tiling(1, Memory.HostCache, Memory.HostMemory, Domain(2, 2, 2), OF)
    val lap = Stencil("lap", 5, Map("in" -> AccessSet(List(Index(-1,0,0), Index(1,0,0), Index(0,-1,0), Index(0,1,0), Index(0,0,0)))))
    val fli = Stencil("fli", 1, Map("lap" -> AccessSet(List(Index(1,0,0), Index(0,0,0)))))
    val flj = Stencil("flj", 1, Map("lap" -> AccessSet(List(Index(0,1,0), Index(0,0,0)))))
    val out = Stencil("out", 5, Map("in" -> AccessSet.center, "wgt" -> AccessSet.center, "fli" -> AccessSet(List(Index(-1,0,0), Index(0,0,0))), "flj" -> AccessSet(List(Index(0,-1,0), Index(0,0,0)))))
    val group1 = StencilGroup(tiling1, List(lap, fli), List("lap", "fli"))
    val group2 = StencilGroup(tiling1, List(flj, out), List("out"))
    val group = StencilGroup(tiling0, List(group1, group2), List("out"))

    // compute evaluation domain
    val performance = group.evaluatePerformance

    // evaluate the group1 and group2
    performance.computeOperations(0) should be (None)
    performance.computeOperations(1) should be (None)
    performance.verticalCommunication(0) should be (Map("in" -> (64 + 4 * 32 + 4 * 4), "lap" -> (64 + 32), "fli" -> (64 + 16)))
    performance.verticalCommunication(1) should be (Map("lap" -> (64 + 32), "fli" -> (64 + 16), "in" -> 64, "wgt" -> 64, "out" -> 64))
    performance.lateralCommunication(0) should be (Map())
    performance.lateralCommunication(1) should be (Map())

    // evaluate the nested group 1 performance
    val performance1 = performance.nestedRequirements(0).get
    performance1.computeOperations(0) should be (Some(5 * (8 * (8 + 2 * 4) + 32)))
    performance1.computeOperations(1) should be (Some(1 * (64 + 16)))
    performance1.verticalCommunication(0) should be (Map("in" -> (8 * (8 + 4 * 4 + 2 * 4 + 4 * 2) + (2 * 16 + 4 * 16 + 2 * 8 * 4)), "lap" -> (8 * (8 + 2 * 4) + 32)))
    performance1.verticalCommunication(1) should be (Map("lap" -> (8 * (8 + 2 * 4)), "fli" -> (64 + 16)))
    performance1.lateralCommunication(0) should be (Map())
    performance1.lateralCommunication(1) should be (Map())

    // evaluate the nested group 1 performance
    val performance2 = performance.nestedRequirements(1).get
    performance2.computeOperations(0) should be (Some(1 * (8 * (8 + 4))))
    performance2.computeOperations(1) should be (Some(5 * (64)))
    performance2.verticalCommunication(0) should be (Map("lap" -> (8 * (8 + 2 * 4)), "flj" -> 8 * (8 + 4)))
    performance2.verticalCommunication(1) should be (Map("fli" -> (8 * (8 + 4)), "flj" -> (8 * (8 + 4)), "in" -> 64, "wgt" -> 64, "out" -> 64))
    performance2.lateralCommunication(0) should be (Map())
    performance2.lateralCommunication(1) should be (Map())
  }

  "a StencilGroup" should "compute the HE>HE bandwidth" in {
    val tiling0 = Tiling(0, Memory.HostMemory, Memory.HostMemory, Domain(4, 4, 4), HEPar)
    val tiling1 = Tiling(1, Memory.HostCache, Memory.HostMemory, Domain(2, 2, 2), HEPar)
    val lap = Stencil("lap", 5, Map("in" -> AccessSet(List(Index(-1,0,0), Index(1,0,0), Index(0,-1,0), Index(0,1,0), Index(0,0,0)))))
    val fli = Stencil("fli", 1, Map("lap" -> AccessSet(List(Index(1,0,0), Index(0,0,0)))))
    val flj = Stencil("flj", 1, Map("lap" -> AccessSet(List(Index(0,1,0), Index(0,0,0)))))
    val out = Stencil("out", 5, Map("in" -> AccessSet.center, "wgt" -> AccessSet.center, "fli" -> AccessSet(List(Index(-1,0,0), Index(0,0,0))), "flj" -> AccessSet(List(Index(0,-1,0), Index(0,0,0)))))
    val group1 = StencilGroup(tiling1, List(lap, fli), List("lap", "fli"))
    val group2 = StencilGroup(tiling1, List(flj, out), List("out"))
    val group = StencilGroup(tiling0, List(group1, group2), List("out"))

    // compute evaluation domain
    val performance = group.evaluatePerformance

    // evaluate the group1 and group2
    performance.computeOperations(0) should be (None)
    performance.computeOperations(1) should be (None)
    performance.verticalCommunication(0) should be (Map("in" -> (64 + 4 * 16), "lap" -> 64, "fli" -> 64))
    performance.verticalCommunication(1) should be (Map("lap" -> (64 + 16), "fli" -> (64 + 16), "in" -> 64, "wgt" -> 64, "out" -> 64))
    performance.lateralCommunication(0) should be (Map("lap" -> List(16), "fli" -> List(16)))
    performance.lateralCommunication(1) should be (Map())

    // evaluate the nested group 1 performance
    val performance1 = performance.nestedRequirements(0).get
    performance1.computeOperations(0) should be (Some(5 * 64))
    performance1.computeOperations(1) should be (Some(1 * 64))
    performance1.verticalCommunication(0) should be (Map("in" -> (8 * (8 + 16)), "lap" -> 64))
    performance1.verticalCommunication(1) should be (Map("lap" -> (8 * (8 + 4)), "fli" -> 64))
    performance1.lateralCommunication(0) should be (Map("lap" -> List(16, 16)))
    performance1.lateralCommunication(1) should be (Map())

    // evaluate the nested group 1 performance
    val performance2 = performance.nestedRequirements(1).get
    performance2.computeOperations(0) should be (Some(1 * 64))
    performance2.computeOperations(1) should be (Some(5 * 64))
    performance2.verticalCommunication(0) should be (Map("lap" -> (8 * (8 + 4)), "flj" -> 64))
    performance2.verticalCommunication(1) should be (Map("fli" -> (8 * (8 + 4)), "flj" -> (8 * (8 + 4)), "in" -> 64, "wgt" -> 64, "out" -> 64))
    performance2.lateralCommunication(0) should be (Map("flj" -> List(16, 16)))
    performance2.lateralCommunication(1) should be (Map())
  }

  "a StencilGroup" should "compute the OF>HE bandwidth" in {
    val tiling0 = Tiling(0, Memory.HostMemory, Memory.HostMemory, Domain(4, 4, 4), OF)
    val tiling1 = Tiling(1, Memory.HostCache, Memory.HostMemory, Domain(2, 2, 2), HEPar)
    val lap = Stencil("lap", 5, Map("in" -> AccessSet(List(Index(-1,0,0), Index(1,0,0), Index(0,-1,0), Index(0,1,0), Index(0,0,0)))))
    val fli = Stencil("fli", 1, Map("lap" -> AccessSet(List(Index(1,0,0), Index(0,0,0)))))
    val flj = Stencil("flj", 1, Map("lap" -> AccessSet(List(Index(0,1,0), Index(0,0,0)))))
    val out = Stencil("out", 5, Map("in" -> AccessSet.center, "wgt" -> AccessSet.center, "fli" -> AccessSet(List(Index(-1,0,0), Index(0,0,0))), "flj" -> AccessSet(List(Index(0,-1,0), Index(0,0,0)))))
    val group1 = StencilGroup(tiling1, List(lap, fli), List("lap", "fli"))
    val group2 = StencilGroup(tiling1, List(flj, out), List("out"))
    val group = StencilGroup(tiling0, List(group1, group2), List("out"))

    // compute evaluation domain
    val performance = group.evaluatePerformance

    // evaluate the group1 and group2
    performance.computeOperations(0) should be (None)
    performance.computeOperations(1) should be (None)
    performance.verticalCommunication(0) should be (Map("in" -> (64 + 4 * 32 + 4 * 4), "lap" -> (64 + 32), "fli" -> (64 + 16)))
    performance.verticalCommunication(1) should be (Map("lap" -> (64 + 32), "fli" -> (64 + 16), "in" -> 64, "wgt" -> 64, "out" -> 64))
    performance.lateralCommunication(0) should be (Map())
    performance.lateralCommunication(1) should be (Map())

    // evaluate the nested group 1 performance
    val performance1 = performance.nestedRequirements(0).get
    performance1.computeOperations(0) should be (Some(5 * (64 + 64)))
    performance1.computeOperations(1) should be (Some(1 * (64 + 16)))
    performance1.verticalCommunication(0) should be (Map("in" -> (8 * (8 + 4 * 4) + (4 * 16 + 8 * 16 + 4 * 8 * 2)), "lap" -> (8 * 8 + 64)))
    performance1.verticalCommunication(1) should be (Map("lap" -> (8 * (8 + 2 * 4)), "fli" -> (64 + 16)))
    performance1.lateralCommunication(0) should be (Map("lap" -> List(0,(12 * (4)))))
    performance1.lateralCommunication(1) should be (Map())

    // evaluate the nested group 1 performance
    val performance2 = performance.nestedRequirements(1).get
    performance2.computeOperations(0) should be (Some(1 * (64 + 16)))
    performance2.computeOperations(1) should be (Some(5 * (64)))
    performance2.verticalCommunication(0) should be (Map("lap" -> (8 * (8 + 4) + 2 * 16), "flj" -> (64 + 16)))
    performance2.verticalCommunication(1) should be (Map("fli" -> (8 * (8 + 4)), "flj" -> (8 * (8 + 4)), "in" -> 64, "wgt" -> 64, "out" -> 64))
    performance2.lateralCommunication(0) should be (Map("flj" -> List(0,(8 * (4)))))
    performance2.lateralCommunication(1) should be (Map())
  }

  "a StencilGroup" should "ccompute the HE>OF bandwidth" in {
    val tiling0 = Tiling(0, Memory.HostMemory, Memory.HostMemory, Domain(4, 4, 4), HEPar)
    val tiling1 = Tiling(1, Memory.HostCache, Memory.HostMemory, Domain(2, 2, 2), OF)
    val lap = Stencil("lap", 5, Map("in" -> AccessSet(List(Index(-1,0,0), Index(1,0,0), Index(0,-1,0), Index(0,1,0), Index(0,0,0)))))
    val fli = Stencil("fli", 1, Map("lap" -> AccessSet(List(Index(1,0,0), Index(0,0,0)))))
    val flj = Stencil("flj", 1, Map("lap" -> AccessSet(List(Index(0,1,0), Index(0,0,0)))))
    val out = Stencil("out", 5, Map("in" -> AccessSet.center, "wgt" -> AccessSet.center, "fli" -> AccessSet(List(Index(-1,0,0), Index(0,0,0))), "flj" -> AccessSet(List(Index(0,-1,0), Index(0,0,0)))))
    val group1 = StencilGroup(tiling1, List(lap, fli), List("lap", "fli"))
    val group2 = StencilGroup(tiling1, List(flj, out), List("out"))
    val group = StencilGroup(tiling0, List(group1, group2), List("out"))

    // compute evaluation domain
    val performance = group.evaluatePerformance

    // evaluate the group1 and group2
    performance.computeOperations(0) should be (None)
    performance.computeOperations(1) should be (None)
    performance.verticalCommunication(0) should be (Map("in" -> (64 + 5 * 16 + 2 * 4), "lap" -> 64, "fli" -> 64))
    performance.verticalCommunication(1) should be (Map("lap" -> (64 + 32), "fli" -> (64 + 16), "in" -> 64, "wgt" -> 64, "out" -> 64))
    performance.lateralCommunication(0) should be (Map("lap" -> List(32), "fli" -> List(16)))
    performance.lateralCommunication(1) should be (Map())

    // evaluate the nested group 1 performance
    val performance1 = performance.nestedRequirements(0).get
    performance1.computeOperations(0) should be (Some(5 * (8 * (8 + 4))))
    performance1.computeOperations(1) should be (Some(1 * (64)))
    performance1.verticalCommunication(0) should be (Map("in" -> (8 * (8 + 3 * 4 + 3 * 4)), "lap" -> (8 * (8 + 4))))
    performance1.verticalCommunication(1) should be (Map("lap" -> (8 * (8 + 4)), "fli" -> (64)))
    performance1.lateralCommunication(0) should be (Map())
    performance1.lateralCommunication(1) should be (Map())

    // evaluate the nested group 1 performance
    val performance2 = performance.nestedRequirements(1).get
    performance2.computeOperations(0) should be (Some(1 * (8 * (8 + 4))))
    performance2.computeOperations(1) should be (Some(5 * (64)))
    performance2.verticalCommunication(0) should be (Map("lap" -> (8 * (8 + 2 * 4)), "flj" -> (8 * (8 + 4))))
    performance2.verticalCommunication(1) should be (Map("fli" -> (8 * (8 + 4)), "flj" -> (8 * (8 + 4)), "in" -> 64, "wgt" -> 64, "out" -> 64))
    performance2.lateralCommunication(0) should be (Map())
    performance2.lateralCommunication(1) should be (Map())
  }
}
