import org.scalatest._

/**
 * test isl bindings
 */
class ISLSpec extends FlatSpec with Matchers {

  "findOrCreateID" should "create a single identifier per name" in {
    val id1 = ISLContext.findOrCreateID("same")
    val id2 = ISLContext.findOrCreateID("same")
    val id3 = ISLContext.findOrCreateID("different")

    id1 should equal (id2)
    id3 should not equal id1
    id3 should not equal id2
  }

  "Barvinok" should "count the points in a set" in {
    val domain = Domain(3, 4, 5)
    Barvinok.count(domain.convertToISLSet.toString) should be(60)
  }

  "a Domain" should "return the corresponding ISL set" in {
    val domain = Domain(3, 4, 5)
    val set = domain.convertToISLSet

    set.dimIsBounded(isl.DimType.Set, 0) should be (true)
    set.dimIsBounded(isl.DimType.Set, 1) should be (true)
    set.dimIsBounded(isl.DimType.Set, 2) should be (true)
    set.countVal.getNum.intValue should be (60)
  }

  "an AccessSet" should "return the corresponding ISL map" in {
    val accesses = AccessSet(List(Index(0,0,0), Index(1,-2,3)))
    val accessMap = accesses.convertToISLMap

    val expected = isl.Map.readFromStr(ISLContext.ctx,
      "{ [i0, i1, i2] -> [1 + i0, -2 + i1, 3 + i2]; [i0, i1, i2] -> [i0, i1, i2] }"
    )
    accessMap.isEqual(expected) should not be 0
  }

  "a Stencil" should "compute a dependency map" in {
    val stencil = Stencil("out", 5, Map("in" -> AccessSet.center, "wgt" -> AccessSet.center, "fli" -> AccessSet(List(Index(-1,0,0), Index(0,0,0))), "flj" -> AccessSet(List(Index(0,-1,0), Index(0,0,0)))))

    val expected = isl.UnionMap.readFromStr(ISLContext.ctx,
      "{ out[i0, i1, i2] -> wgt[i0, i1, i2]; " +
        "out[i0, i1, i2] -> fli[o0, i1, i2] : o0 <= i0 and o0 >= -1 + i0; " +
        "out[i0, i1, i2] -> in[i0, i1, i2]; " +
        "out[i0, i1, i2] -> flj[i0, o1, i2] : o1 <= i1 and o1 >= -1 + i1 }"
    )
    stencil.dependencies.isEqual(expected) should not be 0
  }

  "a StencilGroup" should "compute a dependency map" in {
    val tiling0 = Tiling(0, Memory.HostMemory, Memory.HostMemory, Domain(256, 256, 64), OF)
    val tiling1 = Tiling(1, Memory.HostCache, Memory.HostMemory, Domain(8, 8, 64), OF)
    val lap = Stencil("lap", 5, Map("in" -> AccessSet(List(Index(-1,0,0), Index(1,0,0), Index(0,-1,0), Index(0,1,0), Index(0,0,0)))))
    val fli = Stencil("fli", 1, Map("lap" -> AccessSet(List(Index(1,0,0), Index(0,0,0)))))
    val flj = Stencil("flj", 1, Map("lap" -> AccessSet(List(Index(0,1,0), Index(0,0,0)))))
    val out = Stencil("out", 5, Map("in" -> AccessSet.center, "wgt" -> AccessSet.center, "fli" -> AccessSet(List(Index(-1,0,0), Index(0,0,0))), "flj" -> AccessSet(List(Index(0,-1,0), Index(0,0,0)))))
    val group1 = StencilGroup(tiling1, List(lap, fli), List("lap", "fli"))
    val group2 = StencilGroup(tiling1, List(flj, out), List("out"))
    val group = StencilGroup(tiling0, List(group1, group2), List("out"))

    val expected1 = isl.UnionMap.readFromStr(ISLContext.ctx,
      "{ fli[i0, i1, i2] -> in[o0, o1, i2] : o1 >= -1 + i0 + i1 - o0 and o1 <= 1 - i0 + i1 + o0 and o1 >= -2 - i0 + i1 + o0 and o1 <= 2 + i0 + i1 - o0; " +
        "lap[i0, i1, i2] -> in[o0, o1, i2] : o1 >= -1 + i0 + i1 - o0 and o1 <= 1 - i0 + i1 + o0 and o1 >= -1 - i0 + i1 + o0 and o1 <= 1 + i0 + i1 - o0 }"
    )
    val expected2 = isl.UnionMap.readFromStr(ISLContext.ctx,
      "{ out[i0, i1, i2] -> lap[i0, o1, i2] : o1 >= -1 + i1 and o1 <= 1 + i1; " +
        "out[i0, i1, i2] -> fli[o0, i1, i2] : o0 >= -1 + i0 and o0 <= i0; " +
        "out[i0, i1, i2] -> wgt[i0, i1, i2]; " +
        "out[i0, i1, i2] -> in[i0, i1, i2] }"
    )
    val expected = isl.UnionMap.readFromStr(ISLContext.ctx,
      "{ out[i0, i1, i2] -> wgt[i0, i1, i2]; " +
        "out[i0, i1, i2] -> in[o0, o1, i2] : o1 >= -2 + i0 + i1 - o0 and o1 <= 2 - i0 + i1 + o0 and o1 >= -2 - i0 + i1 + o0 and o1 <= 2 + i0 + i1 - o0 }"
    )

    group1.dependencies.isEqual(expected1) should not be 0
    group2.dependencies.isEqual(expected2) should not be 0
    group.dependencies.isEqual(expected) should not be 0
  }

  "a StencilGroup" should "compute a flatten dependency map" in {
    val tiling0 = Tiling(0, Memory.HostMemory, Memory.HostMemory, Domain(256, 256, 64), OF)
    val tiling1 = Tiling(1, Memory.HostCache, Memory.HostMemory, Domain(8, 8, 64), OF)
    val lap = Stencil("lap", 5, Map("in" -> AccessSet(List(Index(-1,0,0), Index(1,0,0), Index(0,-1,0), Index(0,1,0), Index(0,0,0)))))
    val fli = Stencil("fli", 1, Map("lap" -> AccessSet(List(Index(1,0,0), Index(0,0,0)))))
    val flj = Stencil("flj", 1, Map("lap" -> AccessSet(List(Index(0,1,0), Index(0,0,0)))))
    val out = Stencil("out", 5, Map("in" -> AccessSet.center, "wgt" -> AccessSet.center, "fli" -> AccessSet(List(Index(-1,0,0), Index(0,0,0))), "flj" -> AccessSet(List(Index(0,-1,0), Index(0,0,0)))))
    val group1 = StencilGroup(tiling1, List(lap, fli), List("lap", "fli"))
    val group2 = StencilGroup(tiling1, List(flj, out), List("out"))
    val group = StencilGroup(tiling0, List(group1, group2), List("out"))

    val expected1 = isl.UnionMap.readFromStr(ISLContext.ctx,
      "{ fli[i0, i1, i2] -> lap[o0, i1, i2] : o0 >= i0 and o0 <= 1 + i0; " +
        "fli[i0, i1, i2] -> in[o0, o1, i2] : o1 >= -1 + i0 + i1 - o0 and o1 <= 1 - i0 + i1 + o0 and o1 >= -2 - i0 + i1 + o0 and o1 <= 2 + i0 + i1 - o0; " +
        "lap[i0, i1, i2] -> in[o0, o1, i2] : o1 >= -1 + i0 + i1 - o0 and o1 <= 1 - i0 + i1 + o0 and o1 >= -1 - i0 + i1 + o0 and o1 <= 1 + i0 + i1 - o0 }"
    )
    val expected2 = isl.UnionMap.readFromStr(ISLContext.ctx,
      "{ out[i0, i1, i2] -> wgt[i0, i1, i2]; " +
        "out[i0, i1, i2] -> fli[o0, i1, i2] : o0 >= -1 + i0 and o0 <= i0; " +
        "out[i0, i1, i2] -> flj[i0, o1, i2] : o1 >= -1 + i1 and o1 <= i1; " +
        "out[i0, i1, i2] -> in[i0, i1, i2]; " +
        "out[i0, i1, i2] -> lap[i0, o1, i2] : o1 >= -1 + i1 and o1 <= 1 + i1; " +
        "flj[i0, i1, i2] -> lap[i0, o1, i2] : o1 >= i1 and o1 <= 1 + i1 }"
    )
    val expected = isl.UnionMap.readFromStr(ISLContext.ctx,
      "{ out[i0, i1, i2] -> wgt[i0, i1, i2]; " +
        "out[i0, i1, i2] -> lap[i0, o1, i2] : o1 >= -1 + i1 and o1 <= 1 + i1; " +
        "out[i0, i1, i2] -> lap[o0, i1, i2] : o0 >= -1 + i0 and o0 <= 1 + i0; " +
        "out[i0, i1, i2] -> fli[o0, i1, i2] : o0 >= -1 + i0 and o0 <= i0; " +
        "out[i0, i1, i2] -> in[o0, o1, i2] : o1 >= -2 + i0 + i1 - o0 and o1 <= 2 - i0 + i1 + o0 and o1 >= -2 - i0 + i1 + o0 and o1 <= 2 + i0 + i1 - o0; " +
        "out[i0, i1, i2] -> flj[i0, o1, i2] : o1 >= -1 + i1 and o1 <= i1; " +
        "fli[i0, i1, i2] -> lap[o0, i1, i2] : o0 >= i0 and o0 <= 1 + i0; " +
        "fli[i0, i1, i2] -> in[o0, o1, i2] : o1 >= -1 + i0 + i1 - o0 and o1 <= 1 - i0 + i1 + o0 and o1 >= -2 - i0 + i1 + o0 and o1 <= 2 + i0 + i1 - o0; " +
        "flj[i0, i1, i2] -> lap[i0, o1, i2] : o1 >= i1 and o1 <= 1 + i1; " +
        "flj[i0, i1, i2] -> in[o0, o1, i2] : o1 <= 2 - i0 + i1 + o0 and o1 <= 2 + i0 + i1 - o0 and o1 >= -1 + i0 + i1 - o0 and o1 >= -1 - i0 + i1 + o0; " +
        "lap[i0, i1, i2] -> in[o0, o1, i2] : o1 >= -1 + i0 + i1 - o0 and o1 <= 1 - i0 + i1 + o0 and o1 >= -1 - i0 + i1 + o0 and o1 <= 1 + i0 + i1 - o0 }"
    )

    group1.flattenDependencies.isEqual(expected1) should not be 0
    group2.flattenDependencies.isEqual(expected2) should not be 0
    group.flattenDependencies.isEqual(expected) should not be 0
  }
}
