import org.scalatest._

/**
 * test the tiling logic
 */
class TilingSpec extends FlatSpec with Matchers {

  "a StencilGroup" should "compute an OF tiling" in {
    val tiling0 = Tiling(0, Memory.HostMemory, Memory.HostMemory, Domain(5, 2, 1), OF)
    val lap = Stencil("lap", 5, Map("in" -> AccessSet(List(Index(-1,0,0), Index(1,0,0), Index(0,-1,0), Index(0,1,0), Index(0,0,0)))))
    val fli = Stencil("fli", 1, Map("lap" -> AccessSet(List(Index(1,0,0), Index(0,0,0)))))
    val flj = Stencil("flj", 1, Map("lap" -> AccessSet(List(Index(0,1,0), Index(0,0,0)))))
    val out = Stencil("out", 5, Map("in" -> AccessSet.center, "wgt" -> AccessSet.center, "fli" -> AccessSet(List(Index(-1,0,0), Index(0,0,0))), "flj" -> AccessSet(List(Index(0,-1,0), Index(0,0,0)))))
    val group = StencilGroup(tiling0, List(lap, fli, flj, out), List("out"))

    // compute evaluation domain
    val (stencilTiling, outputTiling) = group.computeTiling

    // assert the output tiling is equivalent to the stencil tiling
    stencilTiling("out").isEqual(outputTiling("out")) should not be 0
    stencilTiling("fli").isEqual(outputTiling("fli")) should not be 0
    stencilTiling("flj").isEqual(outputTiling("flj")) should not be 0
    stencilTiling("lap").isEqual(outputTiling("lap")) should not be 0

    // compute the size of the origin tile
    val originTile = isl.Set.readFromStr(ISLContext.ctx, "{ [0,0,0] }")

    // make sure each out evaluation is assigned to exactly one tile
    val outTiling = stencilTiling("out")
    val outTilingExp = isl.Map.readFromStr(ISLContext.ctx, "{ out[i0, i1, i2] -> [floor(i0/5), floor(i1/2), floor(i2/1)] }" )
    val outTile = originTile.apply(outTiling.reverse())
    val outTileExp = isl.Set.readFromStr(ISLContext.ctx, "{ out[i0, i1, i2]: 0 <= i0 <= 4 and 0 <= i1 <= 1 and i2 = 0 }")

    // compare the expected tiling and count number of elements on fix domain
    outTiling.isEqual(outTilingExp) should not be 0
    outTile.isEqual(outTileExp) should not be 0

    // make sure the flj evaluations are assigned to one base tile
    val fljTiling = stencilTiling("flj")
    val fljTilingExp = isl.Map.readFromStr(ISLContext.ctx,
      "{ flj[i0, i1, i2] -> [floor(i0/5), floor(i1/2), floor(i2/1)]; " +
        "flj[i0, i1, i2] -> [floor(i0/5), floor((i1 + 1)/2), floor(i2/1)] }"
    )
    val fljTile = originTile.apply(fljTiling.reverse())
    val fljTileExp = isl.Set.readFromStr(ISLContext.ctx, "{ flj[i0, i1, i2]: 0 <= i0 <= 4 and -1 <= i1 <= 1 and i2 = 0 }")

    // compare the expected tiling and count number of elements on fix domain
    fljTiling.isEqual(fljTilingExp) should not be 0
    fljTile.isEqual(fljTileExp) should not be 0

    // make sure the lap evaluations are assigned to one base tile
    val lapTiling = stencilTiling("lap")
    val lapTilingExp = isl.Map.readFromStr(ISLContext.ctx,
      "{ lap[i0, i1, i2] -> [floor(i0/5), floor(i1/2), floor(i2/1)]; " +
        "lap[i0, i1, i2] -> [floor(i0/5), floor((i1 + 1)/2), floor(i2/1)]; " +
        "lap[i0, i1, i2] -> [floor(i0/5), floor((i1 - 1)/2), floor(i2/1)]; " +
        "lap[i0, i1, i2] -> [floor((i0 + 1)/5), floor(i1/2), floor(i2/1)]; " +
        "lap[i0, i1, i2] -> [floor((i0 - 1)/5), floor(i1/2), floor(i2/1)] }"
    )
    val lapTile = originTile.apply(lapTiling.reverse())
    val lapTileExp = isl.Set.readFromStr(ISLContext.ctx,
      "{ lap[i0, i1, i2]: (0 <= i0 <= 4 and -1 <= i1 <= 2 and i2 = 0) or (-1 <= i0 <= 5 and 0 <= i1 <= 1 and i2 = 0) }"
    )

    // compare the expected tiling and count number of elements on fix domain
    lapTiling.isEqual(lapTilingExp) should not be 0
    lapTile.isEqual(lapTileExp) should not be 0
  }

  "a StencilGroup" should "compute an HE tiling" in {
    val tiling0 = Tiling(0, Memory.HostMemory, Memory.HostMemory, Domain(5, 2, 1), HEPar)
    val lap = Stencil("lap", 5, Map("in" -> AccessSet(List(Index(-1,0,0), Index(1,0,0), Index(0,-1,0), Index(0,1,0), Index(0,0,0)))))
    val fli = Stencil("fli", 1, Map("lap" -> AccessSet(List(Index(1,0,0), Index(0,0,0)))))
    val flj = Stencil("flj", 1, Map("lap" -> AccessSet(List(Index(0,1,0), Index(0,0,0)))))
    val out = Stencil("out", 5, Map("in" -> AccessSet.center, "wgt" -> AccessSet.center, "fli" -> AccessSet(List(Index(-1,0,0), Index(0,0,0))), "flj" -> AccessSet(List(Index(0,-1,0), Index(0,0,0)))))
    val group = StencilGroup(tiling0, List(lap, fli, flj, out), List("out"))

    // compute evaluation domain
    val (stencilTiling, outputTiling) = group.computeTiling

    // assert the output tiling is equivalent to the stencil tiling
    stencilTiling("out").isEqual(outputTiling("out")) should not be 0
    stencilTiling("fli").isEqual(outputTiling("fli")) should not be 0
    stencilTiling("flj").isEqual(outputTiling("flj")) should not be 0
    stencilTiling("lap").isEqual(outputTiling("lap")) should not be 0

    // compute the size of the origin tile
    val originTile = isl.Set.readFromStr(ISLContext.ctx, "{ [0,0,0] }")

    // make sure each out evaluation is assigned to exactly one tile
    val outTiling = stencilTiling("out")
    val outTilingExp = isl.Map.readFromStr(ISLContext.ctx, "{ out[i0, i1, i2] -> [floor(i0/5), floor(i1/2), floor(i2/1)] }" )
    val outTile = originTile.apply(outTiling.reverse())
    val outTileExp = isl.Set.readFromStr(ISLContext.ctx, "{ out[i0, i1, i2]: 0 <= i0 <= 4 and 0 <= i1 <= 1 and i2 = 0 }")

    // compare the expected tiling and count number of elements on fix domain
    outTiling.isEqual(outTilingExp) should not be 0
    outTile.isEqual(outTileExp) should not be 0

    // make sure the flj evaluations are assigned to one base tile
    val fljTiling = stencilTiling("flj")
    val fljTilingExp = isl.Map.readFromStr(ISLContext.ctx,
      "{ flj[i0, i1, i2] -> [floor(i0/5), floor(i1/2), floor(i2/1)] }"
    )
    val fljTile = originTile.apply(fljTiling.reverse())
    val fljTileExp = isl.Set.readFromStr(ISLContext.ctx, "{ flj[i0, i1, i2]: 0 <= i0 <= 4 and 0 <= i1 <= 1 and i2 = 0 }")

    // compare the expected tiling and count number of elements on fix domain
    fljTiling.isEqual(fljTilingExp) should not be 0
    fljTile.isEqual(fljTileExp) should not be 0

    // make sure the lap evaluations are assigned to one base tile
    val lapTiling = stencilTiling("lap")
    val lapTilingExp = isl.Map.readFromStr(ISLContext.ctx,
      "{ lap[i0, i1, i2] -> [floor(i0/5), floor(i1/2), floor(i2/1)] }"
    )
    val lapTile = originTile.apply(lapTiling.reverse())
    val lapTileExp = isl.Set.readFromStr(ISLContext.ctx, "{ lap[i0, i1, i2]: 0 <= i0 <= 4 and 0 <= i1 <= 1 and i2 = 0 }")

    // compare the expected tiling and count number of elements on fix domain
    lapTiling.isEqual(lapTilingExp) should not be 0
    lapTile.isEqual(lapTileExp) should not be 0
  }

  "a StencilGroup" should "compute an SEQ tiling" in {
    val tiling0 = Tiling(0, Memory.HostMemory, Memory.HostMemory, Domain(5, 2, 3), HEKInc)
    val lap = Stencil("lap", 5, Map("in" -> AccessSet(List(Index(-1,0,0), Index(1,0,0), Index(0,0,-1), Index(0,0,1), Index(0,0,0))))) // exchange j and k
    val fli = Stencil("fli", 1, Map("lap" -> AccessSet(List(Index(1,0,0), Index(0,0,0)))))
    val flj = Stencil("flj", 1, Map("lap" -> AccessSet(List(Index(0,0,1), Index(0,0,0))))) // exchange j and k
    val out = Stencil("out", 5, Map("in" -> AccessSet.center, "wgt" -> AccessSet.center, "fli" -> AccessSet(List(Index(-1,0,0), Index(0,0,0))), "flj" -> AccessSet(List(Index(0,0,-1), Index(0,0,0)))))
    val group = StencilGroup(tiling0, List(lap, fli, flj, out), List("out"))

    // compute evaluation domain
    val (stencilTiling, outputTiling) = group.computeTiling

    // assert the output tiling is equivalent to the stencil tiling
    stencilTiling("out").isEqual(outputTiling("out")) should not be 0
    stencilTiling("fli").isEqual(outputTiling("fli")) should not be 0
    stencilTiling("flj").isEqual(outputTiling("flj")) should not be 0
    stencilTiling("lap").isEqual(outputTiling("lap")) should not be 0

    // compute the size of the origin tile
    val originTile = isl.Set.readFromStr(ISLContext.ctx, "{ [0,0,0] }")

    // make sure each out evaluation is assigned to exactly one tile
    val outTiling = stencilTiling("out")
    val outTilingExp = isl.Map.readFromStr(ISLContext.ctx, "{ out[i0, i1, i2] -> [floor(i0/5), floor(i1/2), floor(i2/3)] }" )
    val outTile = originTile.apply(outTiling.reverse())
    val outTileExp = isl.Set.readFromStr(ISLContext.ctx, "{ out[i0, i1, i2]: 0 <= i0 <= 4 and 0 <= i1 <= 1 and 0 <= i2 <= 2 }")

    // compare the expected tiling and count number of elements on fix domain
    outTiling.isEqual(outTilingExp) should not be 0
    outTile.isEqual(outTileExp) should not be 0

    // make sure the flj evaluations are assigned to one base tile
    val fljTiling = stencilTiling("flj")
    val fljTilingExp = isl.Map.readFromStr(ISLContext.ctx,
      "{ flj[i0, i1, i2] -> [floor(i0/5), floor(i1/2), floor(i2/3)] }"
    )
    val fljTile = originTile.apply(fljTiling.reverse())
    val fljTileExp = isl.Set.readFromStr(ISLContext.ctx, "{ flj[i0, i1, i2]: 0 <= i0 <= 4 and 0 <= i1 <= 1 and 0 <= i2 <= 2 }")

    // compare the expected tiling and count number of elements on fix domain
    fljTiling.isEqual(fljTilingExp) should not be 0
    fljTile.isEqual(fljTileExp) should not be 0

    // make sure the lap evaluations are assigned to one base tile
    val lapTiling = stencilTiling("lap")
    val lapTilingExp = isl.Map.readFromStr(ISLContext.ctx,
      "{ lap[i0, i1, i2] -> [floor(i0/5), floor(i1/2), floor((i2-1)/3)] }"
    )
    val lapTile = originTile.apply(lapTiling.reverse())
    val lapTileExp = isl.Set.readFromStr(ISLContext.ctx, "{ lap[i0, i1, i2]: 0 <= i0 <= 4 and 0 <= i1 <= 1 and 1 <= i2 <= 3 }")

    // compare the expected tiling and count number of elements on fix domain
    lapTiling.isEqual(lapTilingExp) should not be 0
    lapTile.isEqual(lapTileExp) should not be 0
  }

  "a StencilGroup" should "compute an OF>OF tiling" in {
    val tiling0 = Tiling(0, Memory.HostMemory, Memory.HostMemory, Domain(4, 4, 4), OF)
    val tiling1 = Tiling(1, Memory.HostCache, Memory.HostMemory, Domain(2, 2, 2), OF)
    val lap = Stencil("lap", 5, Map("in" -> AccessSet(List(Index(-1,0,0), Index(1,0,0), Index(0,-1,0), Index(0,1,0), Index(0,0,0)))))
    val fli = Stencil("fli", 1, Map("lap" -> AccessSet(List(Index(1,0,0), Index(0,0,0)))))
    val flj = Stencil("flj", 1, Map("lap" -> AccessSet(List(Index(0,1,0), Index(0,0,0)))))
    val out = Stencil("out", 5, Map("in" -> AccessSet.center, "wgt" -> AccessSet.center, "fli" -> AccessSet(List(Index(-1,0,0), Index(0,0,0))), "flj" -> AccessSet(List(Index(0,-1,0), Index(0,0,0)))))
    val group1 = StencilGroup(tiling1, List(lap, fli), List("lap", "fli"))
    val group2 = StencilGroup(tiling1, List(flj, out), List("out"))
    val group = StencilGroup(tiling0, List(group1, group2), List("out"))

    val fliFilter = isl.Set.universe(isl.Space.setAlloc(ISLContext.ctx, 0, 3)).setTupleId(ISLContext.findOrCreateID("fli"))
    val lapFilter = isl.Set.universe(isl.Space.setAlloc(ISLContext.ctx, 0, 3)).setTupleId(ISLContext.findOrCreateID("lap"))

    // compute evaluation domain
    val (stencilTiling, outputTiling) = group.computeTiling

    // assert the output tiling is equivalent to the stencil tiling
    stencilTiling("out").isEqual(outputTiling("out")) should not be 0
    stencilTiling("flj").isEqual(outputTiling("flj")) should not be 0
    stencilTiling("lap").isEqual(outputTiling("lap")) should not be 0

    // make sure lap domain is a subset
    stencilTiling("fli").isEqual(outputTiling("fli").intersectDomain(fliFilter)) should not be 0
    outputTiling("fli").intersectDomain(lapFilter).isSubset(outputTiling("lap")) should not be 0
    outputTiling("fli").intersectDomain(lapFilter).isEqual(outputTiling("lap")) should be (0)

    // compute the size of the origin tile
    val originTile = isl.Set.readFromStr(ISLContext.ctx, "{ [0,0,0,0,0,0] }")
    val applicationDomain = isl.Set.readFromStr(ISLContext.ctx, "{ [o0,o1,o2,o3,o4,o5]: 0 <= o0 <= 1 and 0 <= o1 <= 1 and 0 <= o2 <= 1 }")

    // make sure each out evaluation is assigned to exactly one tile
    val outTiling = stencilTiling("out")
    val outTilingExp = isl.Map.readFromStr(ISLContext.ctx, "{ out[i0, i1, i2] -> [floor(i0/4), floor(i1/4), floor(i2/4), floor(i0/2), floor(i1/2), floor(i2/2)] }")
    val outTile = originTile.apply(outTiling.reverse())
    val outTileExp = isl.Set.readFromStr(ISLContext.ctx, "{ out[i0, i1, i2]: 0 <= i0 <= 1 and 0 <= i1 <= 1 and 0 <= i2 <= 1 }")
    val outEvaluations = outTiling.intersectRange(applicationDomain).range()

    // compare the expected tiling and count number of elements on fix domain
    outTiling.isEqual(outTilingExp) should not be 0
    outTile.isEqual(outTileExp) should not be 0
    Barvinok.count(outEvaluations.toString) should be (8 * 8) // 8 small tiles nested in 8 large ones

    // make sure the flj evaluations are assigned to one base tile
    val fljTiling = stencilTiling("flj")
    val fljTilingExp = isl.Map.readFromStr(ISLContext.ctx,
      "{ flj[i0, i1, i2] -> [floor(i0/4), floor(i1/4), floor(i2/4), floor(i0/2), floor(i1/2), floor(i2/2)];" +
        "flj[i0, i1, i2] -> [floor(i0/4), floor((i1 + 1)/4), floor(i2/4), floor(i0/2), floor((i1 + 1)/2), floor(i2/2)] }"
    )
    val fljTile = originTile.apply(fljTiling.reverse())
    val fljTileExp = isl.Set.readFromStr(ISLContext.ctx, "{ flj[i0, i1, i2]: 0 <= i0 <= 1 and -1 <= i1 <= 1 and 0 <= i2 <= 1 }")
    val fljEvaluations = fljTiling.intersectRange(applicationDomain).range()

    // compare the expected tiling and count number of elements on fix domain
    fljTiling.isEqual(fljTilingExp) should not be 0
    fljTile.isEqual(fljTileExp) should not be 0
    Barvinok.count(fljEvaluations.toString) should be (8 * 8) // 8 small tiles nested in 8 large ones

    // make sure the flj evaluations are assigned to one base tile
    val fliTiling = stencilTiling("fli")
    val fliTilingExp = isl.Map.readFromStr(ISLContext.ctx,
      "{ fli[i0, i1, i2] -> [floor(i0/4), floor(i1/4), floor(i2/4), floor(i0/2), floor(i1/2), floor(i2/2)];" +
        "fli[i0, i1, i2] -> [floor((i0 + 1)/4), floor(i1/4), floor(i2/4), floor(i0/2), floor(i1/2), floor(i2/2)] }"
    )
    val fliTile = originTile.apply(fliTiling.reverse())
    val fliTileExp = isl.Set.readFromStr(ISLContext.ctx, "{ fli[i0, i1, i2]: 0 <= i0 <= 1 and 0 <= i1 <= 1 and 0 <= i2 <= 1 }")
    val fliEvaluations = fliTiling.intersectRange(applicationDomain).range()

    // compare the expected tiling and count number of elements on fix domain
    fliTiling.isEqual(fliTilingExp) should not be 0
    fliTile.isEqual(fliTileExp) should not be 0
    Barvinok.count(fliEvaluations.toString) should be (8 * 8 + 8 * 4) // each of the 8 tiles has a half tile boundary

    // make sure the lap evaluations are assigned to one base tile
    val lapTiling = stencilTiling("lap")
    val lapTilingExp = isl.Map.readFromStr(ISLContext.ctx,
      "{ lap[i0, i1, i2] -> [floor(i0/4), floor(i1/4), floor(i2/4), floor(i0/2), floor(i1/2), floor(i2/2)]; " +
        "lap[i0, i1, i2] -> [floor(i0/4), floor((i1 + 1)/4), floor(i2/4), floor(i0/2), floor(i1/2), floor(i2/2)]; " +
        "lap[i0, i1, i2] -> [floor(i0/4), floor((i1 - 1)/4), floor(i2/4), floor(i0/2), floor(i1/2), floor(i2/2)]; " +
        "lap[i0, i1, i2] -> [floor((i0 + 1)/4), floor(i1/4), floor(i2/4), floor(i0/2), floor(i1/2), floor(i2/2)]; " + // output dependency (additional small tile)
        "lap[i0, i1, i2] -> [floor((i0 - 1)/4), floor(i1/4), floor(i2/4), floor((i0 - 1)/2), floor(i1/2), floor(i2/2)]; " + // internal boundary
        "lap[i0, i1, i2] -> [floor(i0/4), floor(i1/4), floor(i2/4), floor((i0 - 1)/2), floor(i1/2), floor(i2/2)] } " // internal boundary of additional small tile
    )

    val lapTile = originTile.apply(lapTiling.reverse())
    val lapTileExp = isl.Set.readFromStr(ISLContext.ctx,
      "{ lap[i0, i1, i2]: (0 <= i0 <= 2 and 0 <= i1 <= 1 and 0 <= i2 <= 1) }"
    )
    val lapEvaluations = lapTiling.intersectRange(applicationDomain).range()

    // compare the expected tiling and count number of elements on fix domain
    lapTilingExp.isEqual(lapTiling) should not be 0
    lapTile.isEqual(lapTileExp) should not be 0
    Barvinok.count(lapEvaluations.toString) should be (8 * 8 + 12 * 8) // each of the 8 tiles has three half tile boundary
  }

  "a StencilGroup" should "compute an HE>HE tiling" in {
    val tiling0 = Tiling(0, Memory.HostMemory, Memory.HostMemory, Domain(4, 4, 4), HEPar)
    val tiling1 = Tiling(1, Memory.HostCache, Memory.HostMemory, Domain(2, 2, 2), HEPar)
    val lap = Stencil("lap", 5, Map("in" -> AccessSet(List(Index(-1,0,0), Index(1,0,0), Index(0,-1,0), Index(0,1,0), Index(0,0,0)))))
    val fli = Stencil("fli", 1, Map("lap" -> AccessSet(List(Index(1,0,0), Index(0,0,0)))))
    val flj = Stencil("flj", 1, Map("lap" -> AccessSet(List(Index(0,1,0), Index(0,0,0)))))
    val out = Stencil("out", 5, Map("in" -> AccessSet.center, "wgt" -> AccessSet.center, "fli" -> AccessSet(List(Index(-1,0,0), Index(0,0,0))), "flj" -> AccessSet(List(Index(0,-1,0), Index(0,0,0)))))
    val group1 = StencilGroup(tiling1, List(lap, fli), List("lap", "fli"))
    val group2 = StencilGroup(tiling1, List(flj, out), List("out"))
    val group = StencilGroup(tiling0, List(group1, group2), List("out"))

    val fliFilter = isl.Set.universe(isl.Space.setAlloc(ISLContext.ctx, 0, 3)).setTupleId(ISLContext.findOrCreateID("fli"))
    val lapFilter = isl.Set.universe(isl.Space.setAlloc(ISLContext.ctx, 0, 3)).setTupleId(ISLContext.findOrCreateID("lap"))

    // compute evaluation domain
    val (stencilTiling, outputTiling) = group.computeTiling

    // assert the output tiling is equivalent to the stencil tiling
    stencilTiling("out").isEqual(outputTiling("out")) should not be 0
    stencilTiling("flj").isEqual(outputTiling("flj")) should not be 0
    stencilTiling("lap").isEqual(outputTiling("lap")) should not be 0

    stencilTiling("fli").isEqual(outputTiling("fli").intersectDomain(fliFilter)) should not be 0
    outputTiling("fli").intersectDomain(lapFilter).isEqual(outputTiling("lap")) should not be 0

    // compute the size of the origin tile
    val originTile = isl.Set.readFromStr(ISLContext.ctx, "{ [0,0,0,0,0,0] }")
    val applicationDomain = isl.Set.readFromStr(ISLContext.ctx, "{ [o0,o1,o2,o3,o4,o5]: 0 <= o0 <= 1 and 0 <= o1 <= 1 and 0 <= o2 <= 1 }")

    // make sure each out evaluation is assigned to exactly one tile
    val outTiling = stencilTiling("out")
    val outTilingExp = isl.Map.readFromStr(ISLContext.ctx, "{ out[i0, i1, i2] -> [floor(i0/4), floor(i1/4), floor(i2/4), floor(i0/2), floor(i1/2), floor(i2/2)] }")
    val outTile = originTile.apply(outTiling.reverse())
    val outTileExp = isl.Set.readFromStr(ISLContext.ctx, "{ out[i0, i1, i2]: 0 <= i0 <= 1 and 0 <= i1 <= 1 and 0 <= i2 <= 1 }")
    val outEvaluations = outTiling.intersectRange(applicationDomain).range()

    // compare the expected tiling and count number of elements on fix domain
    outTiling.isEqual(outTilingExp) should not be 0
    outTile.isEqual(outTileExp) should not be 0
    Barvinok.count(outEvaluations.toString) should be (8 * 8) // 8 small tiles nested in 8 large ones

    // make sure the flj evaluations are assigned to one base tile
    val fljTiling = stencilTiling("flj")
    val fljTilingExp = isl.Map.readFromStr(ISLContext.ctx,
      "{ flj[i0, i1, i2] -> [floor(i0/4), floor(i1/4), floor(i2/4), floor(i0/2), floor(i1/2), floor(i2/2)] }"
    )
    val fljTile = originTile.apply(fljTiling.reverse())
    val fljTileExp = isl.Set.readFromStr(ISLContext.ctx, "{ flj[i0, i1, i2]: 0 <= i0 <= 1 and 0 <= i1 <= 1 and 0 <= i2 <= 1 }")
    val fljEvaluations = fljTiling.intersectRange(applicationDomain).range()

    // compare the expected tiling and count number of elements on fix domain
    fljTiling.isEqual(fljTilingExp) should not be 0
    fljTile.isEqual(fljTileExp) should not be 0
    Barvinok.count(fljEvaluations.toString) should be (8 * 8) // 8 small tiles nested in 8 large ones

    // make sure the flj evaluations are assigned to one base tile
    val fliTiling = stencilTiling("fli")
    val fliTilingExp = isl.Map.readFromStr(ISLContext.ctx,
      "{ fli[i0, i1, i2] -> [floor(i0/4), floor(i1/4), floor(i2/4), floor(i0/2), floor(i1/2), floor(i2/2)] }"
    )
    val fliTile = originTile.apply(fliTiling.reverse())
    val fliTileExp = isl.Set.readFromStr(ISLContext.ctx, "{ fli[i0, i1, i2]: 0 <= i0 <= 1 and 0 <= i1 <= 1 and 0 <= i2 <= 1 }")
    val fliEvaluations = fliTiling.intersectRange(applicationDomain).range()

    // compare the expected tiling and count number of elements on fix domain
    fliTiling.isEqual(fliTilingExp) should not be 0
    fliTile.isEqual(fliTileExp) should not be 0
    Barvinok.count(fliEvaluations.toString) should be (8 * 8) // 8 small tiles nested in 8 large ones

    // make sure the lap evaluations are assigned to one base tile
    val lapTiling = stencilTiling("lap")
    val lapTilingExp = isl.Map.readFromStr(ISLContext.ctx,
      "{ lap[i0, i1, i2] -> [floor(i0/4), floor(i1/4), floor(i2/4), floor(i0/2), floor(i1/2), floor(i2/2)] }"
    )

    val lapTile = originTile.apply(lapTiling.reverse())
    val lapTileExp = isl.Set.readFromStr(ISLContext.ctx,
      "{ lap[i0, i1, i2]: (0 <= i0 <= 1 and 0 <= i1 <= 1 and 0 <= i2 <= 1) }"
    )
    val lapEvaluations = lapTiling.intersectRange(applicationDomain).range()

    // compare the expected tiling and count number of elements on fix domain
    lapTilingExp.isEqual(lapTiling) should not be 0
    lapTile.isEqual(lapTileExp) should not be 0
    Barvinok.count(lapEvaluations.toString) should be (8 * 8) // 8 small tiles nested in 8 large ones
  }

  "a StencilGroup" should "compute an OF>HE tiling" in {
    val tiling0 = Tiling(0, Memory.HostMemory, Memory.HostMemory, Domain(4, 4, 4), OF)
    val tiling1 = Tiling(1, Memory.HostCache, Memory.HostMemory, Domain(2, 2, 2), HEPar)
    val lap = Stencil("lap", 5, Map("in" -> AccessSet(List(Index(-1,0,0), Index(1,0,0), Index(0,-1,0), Index(0,1,0), Index(0,0,0)))))
    val fli = Stencil("fli", 1, Map("lap" -> AccessSet(List(Index(1,0,0), Index(0,0,0)))))
    val flj = Stencil("flj", 1, Map("lap" -> AccessSet(List(Index(0,1,0), Index(0,0,0)))))
    val out = Stencil("out", 5, Map("in" -> AccessSet.center, "wgt" -> AccessSet.center, "fli" -> AccessSet(List(Index(-1,0,0), Index(0,0,0))), "flj" -> AccessSet(List(Index(0,-1,0), Index(0,0,0)))))
    val group1 = StencilGroup(tiling1, List(lap, fli), List("lap", "fli"))
    val group2 = StencilGroup(tiling1, List(flj, out), List("out"))
    val group = StencilGroup(tiling0, List(group1, group2), List("out"))

    val fliFilter = isl.Set.universe(isl.Space.setAlloc(ISLContext.ctx, 0, 3)).setTupleId(ISLContext.findOrCreateID("fli"))
    val lapFilter = isl.Set.universe(isl.Space.setAlloc(ISLContext.ctx, 0, 3)).setTupleId(ISLContext.findOrCreateID("lap"))

    // compute evaluation domain
    val (stencilTiling, outputTiling) = group.computeTiling

    // assert the output tiling is equivalent to the stencil tiling
    stencilTiling("out").isEqual(outputTiling("out")) should not be 0
    stencilTiling("flj").isEqual(outputTiling("flj")) should not be 0
    stencilTiling("lap").isEqual(outputTiling("lap")) should not be 0

    // make sure lap domain is a subset
    stencilTiling("fli").isEqual(outputTiling("fli").intersectDomain(fliFilter)) should not be 0
    outputTiling("fli").intersectDomain(lapFilter).isSubset(outputTiling("lap")) should not be 0
    outputTiling("fli").intersectDomain(lapFilter).isEqual(outputTiling("lap")) should be (0)

    // compute the size of the origin tile
    val originTile = isl.Set.readFromStr(ISLContext.ctx, "{ [0,0,0,0,0,0] }")
    val applicationDomain = isl.Set.readFromStr(ISLContext.ctx, "{ [o0,o1,o2,o3,o4,o5]: 0 <= o0 <= 1 and 0 <= o1 <= 1 and 0 <= o2 <= 1 }")

    // make sure each out evaluation is assigned to exactly one tile
    val outTiling = stencilTiling("out")
    val outTilingExp = isl.Map.readFromStr(ISLContext.ctx, "{ out[i0, i1, i2] -> [floor(i0/4), floor(i1/4), floor(i2/4), floor(i0/2), floor(i1/2), floor(i2/2)] }")
    val outTile = originTile.apply(outTiling.reverse())
    val outTileExp = isl.Set.readFromStr(ISLContext.ctx, "{ out[i0, i1, i2]: 0 <= i0 <= 1 and 0 <= i1 <= 1 and 0 <= i2 <= 1 }")
    val outEvaluations = outTiling.intersectRange(applicationDomain).range()

    // compare the expected tiling and count number of elements on fix domain
    outTiling.isEqual(outTilingExp) should not be 0
    outTile.isEqual(outTileExp) should not be 0
    Barvinok.count(outEvaluations.toString) should be (8 * 8) // 8 small tiles nested in 8 large ones

    // make sure the flj evaluations are assigned to one base tile
    val fljTiling = stencilTiling("flj")
    val fljTilingExp = isl.Map.readFromStr(ISLContext.ctx,
      "{ flj[i0, i1, i2] -> [floor(i0/4), floor(i1/4), floor(i2/4), floor(i0/2), floor(i1/2), floor(i2/2)];" +
        "flj[i0, i1, i2] -> [floor(i0/4), floor((i1 + 1)/4), floor(i2/4), floor(i0/2), floor(i1/2), floor(i2/2)] }"
    )
    val fljTile = originTile.apply(fljTiling.reverse())
    val fljTileExp = isl.Set.readFromStr(ISLContext.ctx, "{ flj[i0, i1, i2]: 0 <= i0 <= 1 and 0 <= i1 <= 1 and 0 <= i2 <= 1 }")
    val fljEvaluations = fljTiling.intersectRange(applicationDomain).range()

    // compare the expected tiling and count number of elements on fix domain
    fljTiling.isEqual(fljTilingExp) should not be 0
    fljTile.isEqual(fljTileExp) should not be 0
    Barvinok.count(fljEvaluations.toString) should be (8 * 8 + 8 * 4) // 8 small tiles nested in 8 large ones plus 4 boundary tiles for each large tile

    // make sure the flj evaluations are assigned to one base tile
    val fliTiling = stencilTiling("fli")
    val fliTilingExp = isl.Map.readFromStr(ISLContext.ctx,
      "{ fli[i0, i1, i2] -> [floor(i0/4), floor(i1/4), floor(i2/4), floor(i0/2), floor(i1/2), floor(i2/2)];" +
        "fli[i0, i1, i2] -> [floor((i0 + 1)/4), floor(i1/4), floor(i2/4), floor(i0/2), floor(i1/2), floor(i2/2)] }"
    )
    val fliTile = originTile.apply(fliTiling.reverse())
    val fliTileExp = isl.Set.readFromStr(ISLContext.ctx, "{ fli[i0, i1, i2]: 0 <= i0 <= 1 and 0 <= i1 <= 1 and 0 <= i2 <= 1 }")
    val fliEvaluations = fliTiling.intersectRange(applicationDomain).range()

    // compare the expected tiling and count number of elements on fix domain
    fliTiling.isEqual(fliTilingExp) should not be 0
    fliTile.isEqual(fliTileExp) should not be 0
    Barvinok.count(fliEvaluations.toString) should be (8 * 8 + 8 * 4) // each of the 8 tiles has a half tile boundary

    // make sure the lap evaluations are assigned to one base tile
    val lapTiling = stencilTiling("lap")
    val lapTilingExp = isl.Map.readFromStr(ISLContext.ctx,
      "{ lap[i0, i1, i2] -> [floor(i0/4), floor(i1/4), floor(i2/4), floor(i0/2), floor(i1/2), floor(i2/2)]; " +
        "lap[i0, i1, i2] -> [floor(i0/4), floor((i1 + 1)/4), floor(i2/4), floor(i0/2), floor(i1/2), floor(i2/2)]; " +
        "lap[i0, i1, i2] -> [floor(i0/4), floor((i1 - 1)/4), floor(i2/4), floor(i0/2), floor(i1/2), floor(i2/2)]; " +
        "lap[i0, i1, i2] -> [floor((i0 + 1)/4), floor(i1/4), floor(i2/4), floor(i0/2), floor(i1/2), floor(i2/2)]; " +
        "lap[i0, i1, i2] -> [floor((i0 - 1)/4), floor(i1/4), floor(i2/4), floor(i0/2), floor(i1/2), floor(i2/2)] } "
    )

    val lapTile = originTile.apply(lapTiling.reverse())
    val lapTileExp = isl.Set.readFromStr(ISLContext.ctx,
      "{ lap[i0, i1, i2]: (0 <= i0 <= 1 and 0 <= i1 <= 1 and 0 <= i2 <= 1) }"
    )
    val lapEvaluations = lapTiling.intersectRange(applicationDomain).range()

    // compare the expected tiling and count number of elements on fix domain
    lapTilingExp.isEqual(lapTiling) should not be 0
    lapTile.isEqual(lapTileExp) should not be 0
    Barvinok.count(lapEvaluations.toString) should be (8 * 8 + 16 * 8) // each of the 8 tiles has a boundary in each direction
  }

  "a StencilGroup" should "compute an HE>OF tiling" in {
    val tiling0 = Tiling(0, Memory.HostMemory, Memory.HostMemory, Domain(4, 4, 4), HEPar)
    val tiling1 = Tiling(1, Memory.HostCache, Memory.HostMemory, Domain(2, 2, 2), OF)
    val lap = Stencil("lap", 5, Map("in" -> AccessSet(List(Index(-1,0,0), Index(1,0,0), Index(0,-1,0), Index(0,1,0), Index(0,0,0)))))
    val fli = Stencil("fli", 1, Map("lap" -> AccessSet(List(Index(1,0,0), Index(0,0,0)))))
    val flj = Stencil("flj", 1, Map("lap" -> AccessSet(List(Index(0,1,0), Index(0,0,0)))))
    val out = Stencil("out", 5, Map("in" -> AccessSet.center, "wgt" -> AccessSet.center, "fli" -> AccessSet(List(Index(-1,0,0), Index(0,0,0))), "flj" -> AccessSet(List(Index(0,-1,0), Index(0,0,0)))))
    val group1 = StencilGroup(tiling1, List(lap, fli), List("lap", "fli"))
    val group2 = StencilGroup(tiling1, List(flj, out), List("out"))
    val group = StencilGroup(tiling0, List(group1, group2), List("out"))

    val fliFilter = isl.Set.universe(isl.Space.setAlloc(ISLContext.ctx, 0, 3)).setTupleId(ISLContext.findOrCreateID("fli"))
    val lapFilter = isl.Set.universe(isl.Space.setAlloc(ISLContext.ctx, 0, 3)).setTupleId(ISLContext.findOrCreateID("lap"))

    // compute evaluation domain
    val (stencilTiling, outputTiling) = group.computeTiling

    // assert the output tiling is equivalent to the stencil tiling
    stencilTiling("out").isEqual(outputTiling("out")) should not be 0
    stencilTiling("flj").isEqual(outputTiling("flj")) should not be 0
    stencilTiling("lap").isEqual(outputTiling("lap")) should not be 0

    // make sure lap domain is a subset
    stencilTiling("fli").isEqual(outputTiling("fli").intersectDomain(fliFilter)) should not be 0
    outputTiling("fli").intersectDomain(lapFilter).isSubset(outputTiling("lap")) should not be 0
    outputTiling("fli").intersectDomain(lapFilter).isEqual(outputTiling("lap")) should be (0)

    // compute the size of the origin tile
    val originTile = isl.Set.readFromStr(ISLContext.ctx, "{ [0,0,0,0,0,0] }")
    val applicationDomain = isl.Set.readFromStr(ISLContext.ctx, "{ [o0,o1,o2,o3,o4,o5]: 0 <= o0 <= 1 and 0 <= o1 <= 1 and 0 <= o2 <= 1 }")

    // make sure each out evaluation is assigned to exactly one tile
    val outTiling = stencilTiling("out")
    val outTilingExp = isl.Map.readFromStr(ISLContext.ctx, "{ out[i0, i1, i2] -> [floor(i0/4), floor(i1/4), floor(i2/4), floor(i0/2), floor(i1/2), floor(i2/2)] }")
    val outTile = originTile.apply(outTiling.reverse())
    val outTileExp = isl.Set.readFromStr(ISLContext.ctx, "{ out[i0, i1, i2]: 0 <= i0 <= 1 and 0 <= i1 <= 1 and 0 <= i2 <= 1 }")
    val outEvaluations = outTiling.intersectRange(applicationDomain).range()

    // compare the expected tiling and count number of elements on fix domain
    outTiling.isEqual(outTilingExp) should not be 0
    outTile.isEqual(outTileExp) should not be 0
    Barvinok.count(outEvaluations.toString) should be (8 * 8) // 8 small tiles nested in 8 large ones

    // make sure the flj evaluations are assigned to one base tile
    val fljTiling = stencilTiling("flj")
    val fljTilingExp = isl.Map.readFromStr(ISLContext.ctx,
      "{ flj[i0, i1, i2] -> [floor(i0/4), floor(i1/4), floor(i2/4), floor(i0/2), floor(i1/2), floor(i2/2)];" +
        "flj[i0, i1, i2] -> [floor(i0/4), floor((i1+1)/4), floor(i2/4), floor(i0/2), floor((i1+1)/2), floor(i2/2)] }" // both tiles grow
    )
    val fljTile = originTile.apply(fljTiling.reverse())
    val fljTileExp = isl.Set.readFromStr(ISLContext.ctx, "{ flj[i0, i1, i2]: 0 <= i0 <= 1 and -1 <= i1 <= 1 and 0 <= i2 <= 1 }")
    val fljEvaluations = fljTiling.intersectRange(applicationDomain).range()

    // compare the expected tiling and count number of elements on fix domain
    fljTiling.isEqual(fljTilingExp) should not be 0
    fljTile.isEqual(fljTileExp) should not be 0
    Barvinok.count(fljEvaluations.toString) should be (8 * 8) // 8 small tiles nested in 8 large ones

    // make sure the flj evaluations are assigned to one base tile
    val fliTiling = stencilTiling("fli")
    val fliTilingExp = isl.Map.readFromStr(ISLContext.ctx,
      "{ fli[i0, i1, i2] -> [floor(i0/4), floor(i1/4), floor(i2/4), floor(i0/2), floor(i1/2), floor(i2/2)] }"
    )
    val fliTile = originTile.apply(fliTiling.reverse())
    val fliTileExp = isl.Set.readFromStr(ISLContext.ctx, "{ fli[i0, i1, i2]: 0 <= i0 <= 1 and 0 <= i1 <= 1 and 0 <= i2 <= 1 }")
    val fliEvaluations = fliTiling.intersectRange(applicationDomain).range()

    // compare the expected tiling and count number of elements on fix domain
    fliTilingExp.isEqual(fliTiling) should not be 0
    fliTile.isEqual(fliTileExp) should not be 0
    Barvinok.count(fliEvaluations.toString) should be (8 * 8) // 8 small tiles nested in 8 large ones

    // make sure the lap evaluations are assigned to one base tile
    val lapTiling = stencilTiling("lap")
    val lapTilingExp = isl.Map.readFromStr(ISLContext.ctx,
      "{ lap[i0, i1, i2] -> [floor(i0/4), floor(i1/4), floor(i2/4), floor(i0/2), floor(i1/2), floor(i2/2)]; " +
        "lap[i0, i1, i2] -> [floor((i0 - 1)/4), floor(i1/4), floor(i2/4), floor((i0 - 1)/2), floor(i1/2), floor(i2/2)] } "
    )

    val lapTile = originTile.apply(lapTiling.reverse())
    val lapTileExp = isl.Set.readFromStr(ISLContext.ctx,
      "{ lap[i0, i1, i2]: (0 <= i0 <= 2 and 0 <= i1 <= 1 and 0 <= i2 <= 1) }"
    )
    val lapEvaluations = lapTiling.intersectRange(applicationDomain).range()

    // compare the expected tiling and count number of elements on fix domain
    lapTilingExp.isEqual(lapTiling) should not be 0
    lapTile.isEqual(lapTileExp) should not be 0
    Barvinok.count(lapEvaluations.toString) should be (8 * 8) // 8 small tiles nested in 8 large ones
  }

  "a StencilGroup" should "compute an SEQ>HE tiling" in {
    val tiling0 = Tiling(0, Memory.HostMemory, Memory.HostMemory, Domain(4, 4, 4), HEKInc)
    val tiling1 = Tiling(1, Memory.HostCache, Memory.HostMemory, Domain(2, 2, 2), HEPar)
    val lap = Stencil("lap", 5, Map("in" -> AccessSet(List(Index(-1,0,0), Index(1,0,0), Index(0,0,-1), Index(0,0,1), Index(0,0,0))))) // exchange j and k
    val fli = Stencil("fli", 1, Map("lap" -> AccessSet(List(Index(1,0,0), Index(0,0,0)))))
    val flj = Stencil("flj", 1, Map("lap" -> AccessSet(List(Index(0,0,1), Index(0,0,0))))) // exchange j and k
    val out = Stencil("out", 5, Map("in" -> AccessSet.center, "wgt" -> AccessSet.center, "fli" -> AccessSet(List(Index(-1,0,0), Index(0,0,0))), "flj" -> AccessSet(List(Index(0,0,-1), Index(0,0,0)))))
    val group1 = StencilGroup(tiling1, List(lap, fli), List("lap", "fli"))
    val group2 = StencilGroup(tiling1, List(flj, out), List("out"))
    val group = StencilGroup(tiling0, List(group1, group2), List("out"))

    val fliFilter = isl.Set.universe(isl.Space.setAlloc(ISLContext.ctx, 0, 3)).setTupleId(ISLContext.findOrCreateID("fli"))
    val lapFilter = isl.Set.universe(isl.Space.setAlloc(ISLContext.ctx, 0, 3)).setTupleId(ISLContext.findOrCreateID("lap"))

    // compute evaluation domain
    val (stencilTiling, outputTiling) = group.computeTiling

    // assert the output tiling is equivalent to the stencil tiling
    stencilTiling("out").isEqual(outputTiling("out")) should not be 0
    stencilTiling("flj").isEqual(outputTiling("flj")) should not be 0
    stencilTiling("lap").isEqual(outputTiling("lap")) should not be 0

    stencilTiling("fli").isEqual(outputTiling("fli").intersectDomain(fliFilter)) should not be 0
    outputTiling("fli").intersectDomain(lapFilter).isEqual(outputTiling("lap")) should not be 0

    // compute the size of the origin tile
    val originTile = isl.Set.readFromStr(ISLContext.ctx, "{ [0,0,0,0,0,0] }")
    val applicationDomain = isl.Set.readFromStr(ISLContext.ctx, "{ [o0,o1,o2,o3,o4,o5]: 0 <= o0 <= 1 and 0 <= o1 <= 1 and 0 <= o2 <= 1 }")

    // make sure each out evaluation is assigned to exactly one tile
    val outTiling = stencilTiling("out")
    val outTilingExp = isl.Map.readFromStr(ISLContext.ctx, "{ out[i0, i1, i2] -> [floor(i0/4), floor(i1/4), floor(i2/4), floor(i0/2), floor(i1/2), floor(i2/2)] }")
    val outTile = originTile.apply(outTiling.reverse())
    val outTileExp = isl.Set.readFromStr(ISLContext.ctx, "{ out[i0, i1, i2]: 0 <= i0 <= 1 and 0 <= i1 <= 1 and 0 <= i2 <= 1 }")
    val outEvaluations = outTiling.intersectRange(applicationDomain).range()

    // compare the expected tiling and count number of elements on fix domain
    outTiling.isEqual(outTilingExp) should not be 0
    outTile.isEqual(outTileExp) should not be 0
    Barvinok.count(outEvaluations.toString) should be (8 * 8) // 8 small tiles nested in 8 large ones

    // make sure the flj evaluations are assigned to one base tile
    val fljTiling = stencilTiling("flj")
    val fljTilingExp = isl.Map.readFromStr(ISLContext.ctx,
      "{ flj[i0, i1, i2] -> [floor(i0/4), floor(i1/4), floor(i2/4), floor(i0/2), floor(i1/2), floor(i2/2)] }"
    )
    val fljTile = originTile.apply(fljTiling.reverse())
    val fljTileExp = isl.Set.readFromStr(ISLContext.ctx, "{ flj[i0, i1, i2]: 0 <= i0 <= 1 and 0 <= i1 <= 1 and 0 <= i2 <= 1 }")
    val fljEvaluations = fljTiling.intersectRange(applicationDomain).range()

    // compare the expected tiling and count number of elements on fix domain
    fljTiling.isEqual(fljTilingExp) should not be 0
    fljTile.isEqual(fljTileExp) should not be 0
    Barvinok.count(fljEvaluations.toString) should be (8 * 8) // 8 small tiles nested in 8 large ones

    // make sure the flj evaluations are assigned to one base tile
    val fliTiling = stencilTiling("fli")
    val fliTilingExp = isl.Map.readFromStr(ISLContext.ctx,
      "{ fli[i0, i1, i2] -> [floor(i0/4), floor(i1/4), floor(i2/4), floor(i0/2), floor(i1/2), floor(i2/2)] }"
    )
    val fliTile = originTile.apply(fliTiling.reverse())
    val fliTileExp = isl.Set.readFromStr(ISLContext.ctx, "{ fli[i0, i1, i2]: 0 <= i0 <= 1 and 0 <= i1 <= 1 and 0 <= i2 <= 1 }")
    val fliEvaluations = fliTiling.intersectRange(applicationDomain).range()

    // compare the expected tiling and count number of elements on fix domain
    fliTiling.isEqual(fliTilingExp) should not be 0
    fliTile.isEqual(fliTileExp) should not be 0
    Barvinok.count(fliEvaluations.toString) should be (8 * 8) // 8 small tiles nested in 8 large ones

    // make sure the lap evaluations are assigned to one base tile
    val lapTiling = stencilTiling("lap")
    val lapTilingExp = isl.Map.readFromStr(ISLContext.ctx,
      "{ lap[i0, i1, i2] -> [floor(i0/4), floor(i1/4), floor((i2-1)/4), floor(i0/2), floor(i1/2), floor((i2-1)/2)] }"
    )

    val lapTile = originTile.apply(lapTiling.reverse())
    val lapTileExp = isl.Set.readFromStr(ISLContext.ctx,
      "{ lap[i0, i1, i2]: (0 <= i0 <= 1 and 0 <= i1 <= 1 and 1 <= i2 <= 2) }"
    )
    val lapEvaluations = lapTiling.intersectRange(applicationDomain).range()

    // compare the expected tiling and count number of elements on fix domain
    lapTilingExp.isEqual(lapTiling) should not be 0
    lapTile.isEqual(lapTileExp) should not be 0
    Barvinok.count(lapEvaluations.toString) should be (8 * 8) // 8 small tiles nested in 8 large ones
  }

  "a StencilGroup" should "compute an OF>SEQ tiling" in {
    val tiling0 = Tiling(0, Memory.HostMemory, Memory.HostMemory, Domain(4, 4, 4), OF)
    val tiling1 = Tiling(1, Memory.HostCache, Memory.HostMemory, Domain(2, 2, 2), HEKInc)
    val lap = Stencil("lap", 5, Map("in" -> AccessSet(List(Index(-1,0,0), Index(1,0,0), Index(0,0,-1), Index(0,0,1), Index(0,0,0))))) // exchange j and k
    val fli = Stencil("fli", 1, Map("lap" -> AccessSet(List(Index(1,0,0), Index(0,0,0)))))
    val flj = Stencil("flj", 1, Map("lap" -> AccessSet(List(Index(0,0,1), Index(0,0,0))))) // exchange j and k
    val out = Stencil("out", 5, Map("in" -> AccessSet.center, "wgt" -> AccessSet.center, "fli" -> AccessSet(List(Index(-1,0,0), Index(0,0,0))), "flj" -> AccessSet(List(Index(0,0,-1), Index(0,0,0)))))
    val group1 = StencilGroup(tiling1, List(lap, flj), List("lap", "flj"))
    val group2 = StencilGroup(tiling1, List(fli, out), List("out"))
    val group = StencilGroup(tiling0, List(group1, group2), List("out"))

    val fliFilter = isl.Set.universe(isl.Space.setAlloc(ISLContext.ctx, 0, 3)).setTupleId(ISLContext.findOrCreateID("fli"))
    val lapFilter = isl.Set.universe(isl.Space.setAlloc(ISLContext.ctx, 0, 3)).setTupleId(ISLContext.findOrCreateID("lap"))

    // compute evaluation domain
    val (stencilTiling, outputTiling) = group.computeTiling

    // assert the output tiling is equivalent to the stencil tiling
    stencilTiling("out").isEqual(outputTiling("out")) should not be 0
    stencilTiling("fli").isEqual(outputTiling("fli")) should not be 0
    stencilTiling("lap").isEqual(outputTiling("lap")) should not be 0

    // make sure lap domain is a subset
    stencilTiling("fli").isEqual(outputTiling("fli").intersectDomain(fliFilter)) should not be 0
    outputTiling("fli").intersectDomain(lapFilter).isSubset(outputTiling("lap")) should not be 0
    outputTiling("fli").intersectDomain(lapFilter).isEqual(outputTiling("lap")) should be (0)

    // compute the size of the origin tile
    val originTile = isl.Set.readFromStr(ISLContext.ctx, "{ [0,0,0,0,0,0] }")
    val applicationDomain = isl.Set.readFromStr(ISLContext.ctx, "{ [o0,o1,o2,o3,o4,o5]: 0 <= o0 <= 1 and 0 <= o1 <= 1 and 0 <= o2 <= 1 }")

    // make sure each out evaluation is assigned to exactly one tile
    val outTiling = stencilTiling("out")
    val outTilingExp = isl.Map.readFromStr(ISLContext.ctx, "{ out[i0, i1, i2] -> [floor(i0/4), floor(i1/4), floor(i2/4), floor(i0/2), floor(i1/2), floor(i2/2)] }")
    val outTile = originTile.apply(outTiling.reverse())
    val outTileExp = isl.Set.readFromStr(ISLContext.ctx, "{ out[i0, i1, i2]: 0 <= i0 <= 1 and 0 <= i1 <= 1 and 0 <= i2 <= 1 }")
    val outEvaluations = outTiling.intersectRange(applicationDomain).range()

    // compare the expected tiling and count number of elements on fix domain
    outTiling.isEqual(outTilingExp) should not be 0
    outTile.isEqual(outTileExp) should not be 0
    Barvinok.count(outEvaluations.toString) should be (8 * 8) // 8 small tiles nested in 8 large ones

    // make sure the flj evaluations are assigned to one base tile
    val fljTiling = stencilTiling("flj")
    val fljTilingExp = isl.Map.readFromStr(ISLContext.ctx,
      "{ flj[i0, i1, i2] -> [floor(i0/4), floor(i1/4), floor(i2/4), floor(i0/2), floor(i1/2), floor(i2/2)];" +
        "flj[i0, i1, i2] -> [floor(i0/4), floor(i1/4), floor((i2+1)/4), floor(i0/2), floor(i1/2), floor(i2/2)] }"
    )
    val fljTile = originTile.apply(fljTiling.reverse())
    val fljTileExp = isl.Set.readFromStr(ISLContext.ctx, "{ flj[i0, i1, i2]: 0 <= i0 <= 1 and 0 <= i1 <= 1 and 0 <= i2 <= 1 }")
    val fljEvaluations = fljTiling.intersectRange(applicationDomain).range()

    // compare the expected tiling and count number of elements on fix domain
    fljTiling.isEqual(fljTilingExp) should not be 0
    fljTile.isEqual(fljTileExp) should not be 0
    Barvinok.count(fljEvaluations.toString) should be (8 * 8 + 8 * 4) // 8 small tiles nested in 8 large ones plus 4 boundary tiles for each large tile

    // make sure the flj evaluations are assigned to one base tile
    val fliTiling = stencilTiling("fli")
    val fliTilingExp = isl.Map.readFromStr(ISLContext.ctx,
      "{ fli[i0, i1, i2] -> [floor(i0/4), floor(i1/4), floor(i2/4), floor(i0/2), floor(i1/2), floor(i2/2)];" +
        "fli[i0, i1, i2] -> [floor((i0 + 1)/4), floor(i1/4), floor(i2/4), floor(i0/2), floor(i1/2), floor(i2/2)] }"
    )
    val fliTile = originTile.apply(fliTiling.reverse())
    val fliTileExp = isl.Set.readFromStr(ISLContext.ctx, "{ fli[i0, i1, i2]: 0 <= i0 <= 1 and 0 <= i1 <= 1 and 0 <= i2 <= 1 }")
    val fliEvaluations = fliTiling.intersectRange(applicationDomain).range()

    // compare the expected tiling and count number of elements on fix domain
    fliTiling.isEqual(fliTilingExp) should not be 0
    fliTile.isEqual(fliTileExp) should not be 0
    Barvinok.count(fliEvaluations.toString) should be (8 * 8 + 8 * 4) // each of the 8 tiles has a half tile boundary

    // make sure the lap evaluations are assigned to one base tile
    val lapTiling = stencilTiling("lap")
    val lapTilingExp = isl.Map.readFromStr(ISLContext.ctx,
      "{ lap[i0, i1, i2] -> [floor(i0/4), floor(i1/4), floor(i2/4), floor(i0/2), floor(i1/2), floor((i2 - 1)/2)]; " +
        "lap[i0, i1, i2] -> [floor(i0/4), floor(i1/4), floor((i2 + 1)/4), floor(i0/2), floor(i1/2), floor((i2 - 1)/2)]; " +
        "lap[i0, i1, i2] -> [floor(i0/4), floor(i1/4), floor((i2 - 1)/4), floor(i0/2), floor(i1/2), floor((i2 - 1)/2)]; " +
        "lap[i0, i1, i2] -> [floor((i0 + 1)/4), floor(i1/4), floor(i2/4), floor(i0/2), floor(i1/2), floor((i2 - 1)/2)]; " +
        "lap[i0, i1, i2] -> [floor((i0 - 1)/4), floor(i1/4), floor(i2/4), floor(i0/2), floor(i1/2), floor((i2 - 1)/2)] } "
    )

    val lapTile = originTile.apply(lapTiling.reverse())
    val lapTileExp = isl.Set.readFromStr(ISLContext.ctx,
      "{ lap[i0, i1, i2]: (0 <= i0 <= 1 and 0 <= i1 <= 1 and 1 <= i2 <= 2) }"
    )
    val lapEvaluations = lapTiling.intersectRange(applicationDomain).range()

    // compare the expected tiling and count number of elements on fix domain
    lapTilingExp.isEqual(lapTiling) should not be 0
    lapTile.isEqual(lapTileExp) should not be 0
    Barvinok.count(lapEvaluations.toString) should be (8 * 8 + 16 * 8) // each of the 8 tiles has a boundary in each direction
  }
}
